# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
sncosmo: A Python package for supernova cosmology
"""

import os

from astropy.config import ConfigItem, ConfigNamespace

__version__ = "2.10.0"


def test(
    package=None,
    test_path=None,
    args=None,
    plugins=None,
    verbose=False,
    pastebin=None,
    pep8=False,
    pdb=False,
    coverage=False,
    open_files=False,
    **kwargs
):
    """
    Run the tests using py.test. A proper set of arguments is constructed and
    passed to `pytest.main`.

    Parameters
    ----------
    package : str, optional
        The name of a specific package to test, e.g. 'io.fits' or 'utils'.
        If nothing is specified all default tests are run.

    test_path : str, optional
        Specify location to test by path. May be a single file or
        directory. Must be specified absolutely or relative to the
        calling directory.

    args : str, optional
        Additional arguments to be passed to `pytest.main` in the `args`
        keyword argument.

    plugins : list, optional
        Plugins to be passed to `pytest.main` in the `plugins` keyword
        argument.

    verbose : bool, optional
        Convenience option to turn on verbose output from py.test. Passing
        True is the same as specifying `-v` in `args`.

    pastebin : {'failed','all',None}, optional
        Convenience option for turning on py.test pastebin output. Set to
        'failed' to upload info for failed tests, or 'all' to upload info
        for all tests.

    pep8 : bool, optional
        Turn on PEP8 checking via the pytest-pep8 plugin and disable normal
        tests. Same as specifying `--pep8 -k pep8` in `args`.

    pdb : bool, optional
        Turn on PDB post-mortem analysis for failing tests. Same as
        specifying `--pdb` in `args`.

    coverage : bool, optional
        Generate a test coverage report.  The result will be placed in
        the directory htmlcov.

    open_files : bool, optional
        Fail when any tests leave files open.  Off by default, because
        this adds extra run time to the test suite.  Works only on
        platforms with a working ``lsof`` command.

    parallel : int, optional
        When provided, run the tests in parallel on the specified
        number of CPUs.  If parallel is negative, it will use the all
        the cores on the machine.  Requires the
        `pytest-xdist <https://pypi.python.org/pypi/pytest-xdist>`_ plugin
        installed. Only available when using Astropy 0.3 or later.

    kwargs
        Any additional keywords passed into this function will be passed
        on to the astropy test runner.  This allows use of test-related
        functionality implemented in later versions of astropy without
        explicitly updating the package template.

    See Also
    --------
    pytest.main : py.test function wrapped by `run_tests`.

    """
    import os
    from astropy.tests.helper import TestRunner

    runner = TestRunner(os.path.dirname(__file__))

    return runner.run_tests(
        package=package,
        test_path=test_path,
        args=args,
        plugins=plugins,
        verbose=verbose,
        pastebin=pastebin,
        pep8=pep8,
        pdb=pdb,
        coverage=coverage,
        open_files=open_files,
        **kwargs
    )


# Create default configurations. The file sncosmo.cfg should be
# kept in sync with the ConfigItems here.
class _Conf(ConfigNamespace):
    """Configuration parameters for sncosmo."""

    data_dir = ConfigItem(
        None,
        "Directory where sncosmo will store and read downloaded data "
        "resources. If None, ASTROPY_CACHE_DIR/sncosmo is created and "
        "used. Example: data_dir = /home/user/data/sncosmo",
        cfgtype="string(default=None)",
    )
    sfd98_dir = ConfigItem(
        None,
        "Directory containing SFD (1998) dust maps, with names: "
        "'SFD_dust_4096_ngp.fits' and 'SFD_dust_4096_sgp.fits'. "
        "Example: sfd98_dir = /home/user/data/sfd98",
        cfgtype="string(default=None)",
    )
    remote_timeout = ConfigItem(10.0, "Remote timeout in seconds.")


# Create an instance of the class we just defined.
# This needs to be done before the imports below because `conf` is used
# in some parts of the library.
conf = _Conf()

# clean up namespace
del os, ConfigItem, ConfigNamespace

# import all the things into the top-level namespace
from .bandpasses import *
from .magsystems import *
from .spectrum import *
from .models import *
from .io import *
from .snanaio import *
from .fitting import *
from .simulation import *
from .plotting import *
from .photdata import *
from .registry import *
from .specmodel import *

# Register all the built-ins.
from .builtins import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import copy
import math
import time
import warnings
from collections import OrderedDict

import numpy as np

from .bandpasses import Bandpass
from .photdata import photometric_data
from .utils import Interp1D, Result, ppf

__all__ = ['fit_lc', 'nest_lc', 'mcmc_lc', 'flatten_result', 'chisq']


class DataQualityError(Exception):
    pass


def generate_chisq(data, model, spectra, signature='iminuit', modelcov=False):
    """Define and return a chisq function for use in optimization.

    This function pre-computes and saves the inverse covariance matrix,
    making subsequent evaluations faster. The model covariance (if specified)
    is fixed at the time the chisq function is generated."""

    # precompute inverse covariance matrix
    if data is not None:
        cov = (np.diag(data.fluxerr**2) if data.fluxcov is None else
               data.fluxcov)
        if modelcov:
            _, mcov = model.bandfluxcov(data.band, data.time,
                                        zp=data.zp, zpsys=data.zpsys)
            cov = cov + mcov
        invcov = np.linalg.pinv(cov)

    # If we have spectra, build the covariance matrix for them individually.
    if spectra is not None:
        if modelcov:
            raise ValueError('modelcov not supported for spectra')

        spectra = np.atleast_1d(spectra)
        spectra_invcovs = []
        for spectrum in spectra:
            spectra_invcovs.append(np.linalg.pinv(spectrum.fluxcov))

    # iminuit expects each parameter to be a separate argument (including fixed
    # parameters)
    if signature == 'iminuit':
        def chisq(*parameters):
            model.parameters = parameters

            full_chisq = 0.

            if data is not None:
                model_flux = model.bandflux(data.band, data.time,
                                            zp=data.zp, zpsys=data.zpsys)
                diff = data.flux - model_flux
                phot_chisq = np.dot(np.dot(diff, invcov), diff)
                full_chisq += phot_chisq

            if spectra is not None:
                for spectrum, spec_invcov in zip(spectra, spectra_invcovs):
                    sample_wave, sampling_matrix = \
                        spectrum.get_sampling_matrix()
                    sample_flux = model.flux(spectrum.time, sample_wave)
                    spec_model_flux = (
                        sampling_matrix.dot(sample_flux) /
                        sampling_matrix.dot(np.ones_like(sample_flux))
                    )
                    spec_diff = spectrum.flux - spec_model_flux
                    spec_chisq = spec_invcov.dot(spec_diff).dot(spec_diff)

                    full_chisq += spec_chisq

            return full_chisq
    else:
        raise ValueError("unknown signature: {!r}".format(signature))

    return chisq


def chisq(data, model, modelcov=False):
    """Calculate chisq statistic for the model, given the data.

    Parameters
    ----------
    model : `~sncosmo.Model`
    data : `~astropy.table.Table` or `~numpy.ndarray` or `dict`
        Table of photometric data. Must include certain columns.
        See the "Photometric Data" section of the documentation for
        required columns.
    modelcov : bool
        Include model covariance? Calls ``model.bandfluxcov`` method
        instead of ``model.bandflux``. The source in the model must therefore
        implement covariance.

    Returns
    -------
    chisq : float
    """
    data = photometric_data(data)
    data.sort_by_time()

    if data.fluxcov is None and not modelcov:
        mflux = model.bandflux(data.band, data.time,
                               zp=data.zp, zpsys=data.zpsys)
        return np.sum(((data.flux - mflux) / data.fluxerr)**2)

    else:
        # need to invert a covariance matrix
        cov = (np.diag(data.fluxerr**2) if data.fluxcov is None
               else data.fluxcov)
        if modelcov:
            mflux, mcov = model.bandfluxcov(data.band, data.time,
                                            zp=data.zp, zpsys=data.zpsys)
            cov = cov + mcov
        else:
            mflux = model.bandflux(data.band, data.time,
                                   zp=data.zp, zpsys=data.zpsys)
        invcov = np.linalg.pinv(cov)
        diff = data.flux - mflux
        return np.dot(np.dot(diff, invcov), diff)


def flatten_result(res):
    """Turn a result from fit_lc into a simple dictionary of key, value pairs.

    Useful when saving results to a text file table, where structures
    like a covariance matrix cannot be easily written to a single
    table row.

    Parameters
    ----------
    res : Result
        Result object from `~sncosmo.fit_lc`.

    Returns
    -------
    flatres : Result
        Flattened result. Keys are all strings, values are one of: float, int,
        string), suitable for saving to a text file.
    """

    flat = Result(success=(1 if res.success else 0),
                  ncall=res.ncall,
                  chisq=res.chisq,
                  ndof=res.ndof)

    # Parameters and uncertainties
    for i, n in enumerate(res.param_names):
        flat[n] = res.parameters[i]
        if res.errors is None:
            flat[n + '_err'] = float('nan')
        else:
            flat[n + '_err'] = res.errors.get(n, 0.)

    # Covariances.
    for n1 in res.param_names:
        for n2 in res.param_names:
            key = n1 + '_' + n2 + '_cov'
            if n1 not in res.vparam_names or n2 not in res.vparam_names:
                flat[key] = 0.
            elif res.covariance is None:
                flat[key] = float('nan')
            else:
                i = res.vparam_names.index(n1)
                j = res.vparam_names.index(n2)
                flat[key] = res.covariance[i, j]

    return flat


def _mask_bands(data, model, z_bounds=None):
    if z_bounds is None:
        return model.bandoverlap(data.band)
    else:
        return np.all(model.bandoverlap(data.band, z=z_bounds), axis=1)


def _warn_dropped_bands(data, mask):
    """Warn that we are dropping some bands from the data:"""
    drop_bands = [(b.name if b.name is not None else repr(b))
                  for b in set(data.band[np.invert(mask)])]
    warnings.warn("Dropping following bands from data: " +
                  ", ".join(drop_bands) +
                  "(out of model wavelength range)", RuntimeWarning)


def cut_bands(data, model, z_bounds=None, warn=True):
    mask = _mask_bands(data, model, z_bounds=z_bounds)

    if not np.all(mask):

        # Fail if there are no overlapping bands whatsoever.
        if not np.any(mask):
            raise RuntimeError('No bands in data overlap the model.')

        if warn:
            _warn_dropped_bands(data, mask)

        data = data[mask]

    return data, mask


def t0_bounds(data, model, spectra=None):
    """Determine bounds on t0 parameter of the model.

    The lower bound is such that the latest model time is equal to the
    earliest data time. The upper bound is such that the earliest
    model time is equal to the latest data time."""

    times = []

    if data is not None:
        times.append(data.time)

    if spectra is not None:
        for spec in np.atleast_1d(spectra):
            times.append(spec.time)

    times = np.hstack(times)

    return (model.get('t0') + np.min(times) - model.maxtime(),
            model.get('t0') + np.max(times) - model.mintime())


def _guess_t0_and_amplitude_photometry(data, model, minsnr):
    """Guess t0 and amplitude of the model from photometry."""

    # get data above the signal-to-noise ratio cut
    significant_data = data[(data.flux / data.fluxerr) > minsnr]
    if len(significant_data) == 0:
        raise DataQualityError('No data points with S/N > {0}. Initial '
                               'guessing failed.'.format(minsnr))

    # grid on which to evaluate model light curve
    timegrid = np.linspace(model.mintime(), model.maxtime(),
                           int(model.maxtime() - model.mintime() + 1))

    # get data flux on a consistent scale in order to compare to model
    # flux light curve.
    norm_flux = significant_data.normalized_flux(zp=25., zpsys='ab')

    model_lc = {}
    data_flux = {}
    data_time = {}
    for band in set(significant_data.band):
        model_lc[band] = (
            model.bandflux(band, timegrid, zp=25., zpsys='ab') /
            model.parameters[2])
        mask = significant_data.band == band
        data_flux[band] = norm_flux[mask]
        data_time[band] = significant_data.time[mask]

    if len(model_lc) == 0:
        raise DataQualityError('No data points with S/N > {0}. Initial '
                               'guessing failed.'.format(minsnr))

    # find band with biggest ratio of maximum data flux to maximum model flux
    maxratio = float("-inf")
    maxband = None
    for band in model_lc:
        ratio = np.max(data_flux[band]) / np.max(model_lc[band])
        if ratio > maxratio:
            maxratio = ratio
            maxband = band

    # amplitude guess is the largest ratio
    amplitude = abs(maxratio)

    # time guess is time of max in the band with the biggest ratio
    data_tmax = data_time[maxband][np.argmax(data_flux[maxband])]
    model_tmax = timegrid[np.argmax(model_lc[maxband])]
    t0 = model.get('t0') + data_tmax - model_tmax

    return t0, amplitude


def _guess_t0_and_amplitude_spectra(spectra, model, minsnr):
    """Guess t0 and amplitude of the model from spectra.

    The spectra don't necessarily have the same binning which makes this
    challenging. To handle this, we synthesize photometry in a range of
    different synthetic filters. We then call
    `_guess_t0_and_amplitude_photometry` on this photometry.
    """

    # Build a set of bands to use for synthetic photometry.
    target_band_width = 500     # Angstroms
    minwave = model.minwave()
    maxwave = model.maxwave()
    band_count = int(math.ceil((maxwave - minwave) / target_band_width))
    band_edges = np.linspace(minwave, maxwave, band_count+1)
    band_starts = band_edges[:-1]
    band_ends = band_edges[1:]

    bandpasses = np.array([Bandpass([start, end], [1., 1.]) for start, end
                           in zip(band_starts, band_ends)])

    all_bands = []
    all_fluxes = []
    all_fluxerrs = []
    all_times = []

    for spectrum in np.atleast_1d(spectra):
        # Find the bandpasses that overlap this spectrum.
        band_mask = ((spectrum.bin_edges[0] <= band_starts)
                     & (spectrum.bin_edges[-1] >= band_ends))
        spec_bands = bandpasses[band_mask]

        spec_flux, spec_fluxcov = spectrum.bandfluxcov(spec_bands, zp=25.,
                                                       zpsys='ab')
        spec_fluxerr = np.sqrt(np.diag(spec_fluxcov))

        all_bands.extend(spec_bands)
        all_fluxes.extend(spec_flux)
        all_fluxerrs.extend(spec_fluxerr)
        all_times.extend(np.ones_like(spec_flux) * spectrum.time)

    photometry = photometric_data({
        'band': all_bands,
        'flux': all_fluxes,
        'fluxerr': all_fluxerrs,
        'time': all_times,
        'zp': [25.] * len(all_fluxes),
        'zpsys': ['ab'] * len(all_fluxes),
    })

    # Sort the photometry by time
    photometry = photometry[np.argsort(photometry.time)]

    return _guess_t0_and_amplitude_photometry(photometry, model, minsnr)


def guess_t0_and_amplitude(data, model, minsnr, spectra=None):
    """Guess t0 and amplitude of the model based on the data.

    If we have photometry, we use it for the guessing. If not, we use all
    available spectra instead.
    """
    if data is not None:
        return _guess_t0_and_amplitude_photometry(data, model, minsnr)
    elif spectra is not None:
        return _guess_t0_and_amplitude_spectra(spectra, model, minsnr)
    else:
        raise ValueError('need either photometry or spectra to guess t0 and '
                         'amplitude.')


def _print_iminuit_params(names, fixed_parameters, start_values,
                          start_errors, bounds):
    """Verbose printing of parameters to pass to Minuit"""
    for name in names:
        if name in fixed_parameters:
            continue
        print(name, start_values[name], 'step=', start_errors[name], end=" ")
        if name in bounds:
            print("bounds=", bounds[name], end=" ")
        print()


def _phase_and_wave_mask(data, t0, z, phase_range, wave_range):
    """Return a mask for the data based on an allowed rest-frame phase and/or
    wavelength range (given some t0 and z)."""
    if phase_range is not None:
        data_phase = (data.time - t0) / (1.0 + z)
        phase_mask = ((data_phase > phase_range[0]) &
                      (data_phase < phase_range[1]))

    if wave_range is not None:
        data_obswave = np.array([b.wave_eff for b in data.band])
        data_restwave = data_obswave / (1.0 + z)
        wave_mask = ((data_restwave > wave_range[0]) &
                     (data_restwave < wave_range[1]))

    if phase_range:
        if wave_range:
            return phase_mask & wave_mask
        return phase_mask
    if wave_range:
        return wave_mask
    return None


def _run_iminuit(chisq, parameter_names, start_values, start_errors, bounds,
                 fixed_parameters, maxcall, verbose=False):
    try:
        import iminuit
    except ImportError:
        raise ValueError("Minimization method 'minuit' requires the iminuit "
                         "package")

    # The iminuit API changed significantly in version 2. Handle both the new
    # and old APIs.
    from distutils.version import LooseVersion
    iminuit_version = LooseVersion(iminuit.__version__)

    if verbose:
        print("Initial parameters:")
        _print_iminuit_params(parameter_names, fixed_parameters, start_values,
                              start_errors, bounds)
        print()

    if iminuit_version >= LooseVersion("2.0.0"):
        # New API
        m = iminuit.Minuit(chisq, name=parameter_names, **start_values)

        m.errordef = iminuit.Minuit.LEAST_SQUARES
        m.print_level = 1 if verbose >= 2 else 0
        m.throw_nan = True

        for key, value in start_errors.items():
            m.errors[key] = value

        for key, value in bounds.items():
            m.limits[key] = value

        for key in fixed_parameters:
            m.fixed[key] = True
            m.errors[key] = 0.

        m.migrad(ncall=maxcall)
        fmin = m.fmin
    else:
        # Old API

        # Set up keyword arguments to pass to Minuit initializer.
        kwargs = {}

        # forced_parameters/name keyword also changed in v1.4.3, handle that.
        if iminuit_version >= LooseVersion("1.4.3"):
            kwargs['name'] = parameter_names
        else:
            kwargs['forced_parameters'] = parameter_names

        for key, value in start_values.items():
            kwargs[key] = value

        for key, value in start_errors.items():
            kwargs['error_' + key] = value

        for key, value in bounds.items():
            kwargs['limit_' + key] = value

        for key in fixed_parameters:
            kwargs['fix_' + key] = True
            kwargs['error_' + key] = 0.

        m = iminuit.Minuit(chisq, errordef=1.,
                           print_level=(1 if verbose >= 2 else 0),
                           throw_nan=True, **kwargs)

        fmin, params = m.migrad(ncall=maxcall)

    return m, fmin


def fit_lc(data=None, model=None, vparam_names=[], bounds=None, spectra=None,
           method='minuit', guess_amplitude=True, guess_t0=True, guess_z=True,
           minsnr=5.0, modelcov=False, verbose=False, maxcall=10000,
           phase_range=None, wave_range=None, warn=True):
    """Fit model parameters to data by minimizing chi^2.

    Ths function defines a chi^2 to minimize, makes initial guesses for
    t0 and amplitude, then runs a minimizer.

    Parameters
    ----------
    data : `~astropy.table.Table` or `~numpy.ndarray` or `dict`
        Table of photometric data. Must include certain columns.
        See the "Photometric Data" section of the documentation for
        required columns.
    model : `~sncosmo.Model`
        The model to fit.
    vparam_names : list
        Model parameters to vary in the fit.
    bounds : `dict`, optional
        Bounded range for each parameter. Keys should be parameter
        names, values are tuples. If a bound is not given for some
        parameter, the parameter is unbounded. The exception is
        ``t0``: by default, the minimum bound is such that the latest
        phase of the model lines up with the earliest data point and
        the maximum bound is such that the earliest phase of the model
        lines up with the latest data point.
    spectra : `~sncosmo.Spectrum` or list of `~sncosmo.Spectrum` objects
        A list of spectra to include in the fit.
    guess_amplitude : bool, optional
        Whether or not to guess the amplitude from the data. If false, the
        current model amplitude is taken as the initial value. Only has an
        effect when fitting amplitude. Default is True.
    guess_t0 : bool, optional
        Whether or not to guess t0. Only has an effect when fitting t0.
        Default is True.
    guess_z : bool, optional
        Whether or not to guess z (redshift). Only has an effect when fitting
        redshift. Default is True.
    minsnr : float, optional
        When guessing amplitude and t0, only use data with signal-to-noise
        ratio (flux / fluxerr) greater than this value. Default is 5.
    method : {'minuit'}, optional
        Minimization method to use. Currently there is only one choice.
    modelcov : bool, optional
        Include model covariance when calculating chisq. Default is False.
        If true, the fit is performed multiple times until convergence.
    phase_range : (float, float), optional
        If given, discard data outside this range of phases. Note that
        **the definition of phase varies between models**: For example,
        phase=0 refers to explosion time in some models and time of peak
        B band flux in others.

        *New in version 1.5.0*

    wave_range : (float, float), optional
        If given, discard data with bandpass effective wavelengths outside
        this range.

        *New in version 1.5.0*

    verbose : bool, optional
        Print messages during fitting.
    warn : bool, optional
        Issue a warning when dropping bands outside the wavelength range of
        the model. Default is True.

        *New in version 1.5.0*

    Returns
    -------
    res : Result
        The optimization result represented as a ``Result`` object, which is
        a `dict` subclass with attribute access. Therefore, ``res.keys()``
        provides a list of the attributes. Attributes are:

        - ``success``: boolean describing whether fit succeeded.
        - ``message``: string with more information about exit status.
        - ``ncall``: number of function evaluations.
        - ``chisq``: minimum chi^2 value.
        - ``ndof``: number of degrees of freedom
          (len(data) - len(vparam_names)).
        - ``param_names``: same as ``model.param_names``.
        - ``parameters``: 1-d `~numpy.ndarray` of best-fit values
          (including fixed parameters) corresponding to ``param_names``.
        - ``vparam_names``: list of varied parameter names.
        - ``covariance``: 2-d `~numpy.ndarray` of parameter covariance;
          indicies correspond to order of ``vparam_names``.
        - ``errors``: OrderedDict of varied parameter uncertainties.
          Corresponds to square root of diagonal entries in covariance matrix.
        - ``nfit``: number of times the fit was performed. Can be greater than
          one when model covariance, phase range or wavelength range is used.
          *New in version 1.5.0.*
        - ``data_mask``: Boolean array the same length as data specifying
          whether each observation was used in the final fit.
          *New in version 1.5.0.*

    fitmodel : `~sncosmo.Model`
        A copy of the model with parameters set to best-fit values.

    Notes
    -----

    **t0 guess:** If ``t0`` is being fit and ``guess_t0=True``, the
    function will guess the initial starting point for ``t0`` based on
    the data. The guess is made as follows:

    * Evaluate the time and value of peak flux for the model in each band
      given the current model parameters.
    * Determine the data point with maximum flux in each band, for points
      with signal-to-noise ratio > ``minsnr`` (default is 5). If no points
      meet this criteria, the band is ignored (for the purpose of guessing
      only).
    * For each band, compare model's peak flux to the peak data point. Choose
      the band with the highest ratio of data / model.
    * Set ``t0`` so that the model's time of peak in the chosen band
      corresponds to the peak data point in this band.

    **amplitude guess:** If amplitude (assumed to be the first model parameter)
    is being fit and ``guess_amplitude=True``, the function will guess the
    initial starting point for the amplitude based on the data.

    **redshift guess:** If redshift (``z``) is being fit and ``guess_z=True``,
    the function will set the initial value of ``z`` to the average of the
    bounds on ``z``.

    Examples
    --------

    The `~sncosmo.flatten_result` function can be used to make the result
    a dictionary suitable for appending as rows of a table:

    >>> from astropy.table import Table
    >>> table_rows = []
    >>> for sn in sne:
    ...     res, fitmodel = sncosmo.fit_lc(
    ...          sn, model, ['t0', 'x0', 'x1', 'c'])
    ...     table_rows.append(flatten_result(res))
    >>> t = Table(table_rows)

    """

    if data is not None:
        # Standardize data
        data = photometric_data(data)

        # sort by time (some sources require this)
        # We keep track of indicies that sort the array because we want to be
        # able to report the indexes of the original data that were used in the
        # fit.
        if not np.all(np.ediff1d(data.time) >= 0.0):
            sortidx = np.argsort(data.time)
            data = data[sortidx]
        else:
            sortidx = None
    else:
        sortidx = None

    if spectra is not None:
        # Make sure that we have times for all of the spectra
        for spectrum in np.atleast_1d(spectra):
            if spectrum.time is None:
                raise TypeError("'time' must be set for each spectrum")

    if model is None:
        raise TypeError("missing required argument 'model'")

    # Make a copy of the model so we can modify it with impunity.
    model = copy.copy(model)

    # Check that vparam_names isn't empty and contains only parameters
    # known to the model.
    if len(vparam_names) == 0:
        raise ValueError("no parameters supplied")
    for s in vparam_names:
        if s not in model.param_names:
            raise ValueError("Parameter not in model: " + repr(s))

    # Order vparam_names the same way it is ordered in the model:
    vparam_names = [s for s in model.param_names if s in vparam_names]

    # initialize bounds
    bounds = copy.deepcopy(bounds) if bounds else {}

    # Check that 'z' is bounded (if it is going to be fit).
    if 'z' in vparam_names:
        if 'z' not in bounds or None in bounds['z']:
            raise ValueError('z must be bounded if fit.')
        if guess_z:
            model.set(z=sum(bounds['z']) / 2.)
        if model.get('z') < bounds['z'][0] or model.get('z') > bounds['z'][1]:
            raise ValueError('z out of range.')

    if data is not None:
        # Cut bands that are not allowed by the wavelength range of the model.
        fitdata, support_mask = cut_bands(data, model,
                                          z_bounds=bounds.get('z', None),
                                          warn=warn)
        # Initially this is the complete mask on data.
        data_mask = support_mask
    else:
        fitdata = None
        support_mask = None
        data_mask = None

    # Find t0 bounds to use, if not explicitly given
    if 't0' in vparam_names and 't0' not in bounds:
        bounds['t0'] = t0_bounds(fitdata, model, spectra)

    # Note that in the parameter guessing below, we assume that the source
    # amplitude is the 3rd parameter of the Model (1st parameter of the Source)

    # Turn off guessing if we're not fitting the parameter.
    if model.param_names[2] not in vparam_names:
        guess_amplitude = False
    if 't0' not in vparam_names:
        guess_t0 = False

    # Make guesses for t0 and amplitude.
    # (For now, we assume it is the 3rd parameter of the model.)
    if (guess_amplitude or guess_t0):
        t0, amplitude = guess_t0_and_amplitude(fitdata, model, minsnr, spectra)
        if guess_amplitude:
            model.parameters[2] = amplitude
        if guess_t0:
            model.set(t0=t0)

    if method == 'minuit':
        fixed_parameters = []
        start_values = {}
        start_errors = {}

        # Figure out parameters to pass to iminuit.
        for name in model.param_names:
            start_values[name] = model.get(name)

            # Fix parameters not being varied in the fit.
            if name not in vparam_names:
                fixed_parameters.append(name)
                continue

            # Check that the bounds are valid.
            if name in bounds:
                if None in bounds[name]:
                    raise ValueError('one-sided bounds not allowed for '
                                     'minuit minimizer')

            # Initial step size
            if name in bounds:
                step = 0.02 * (bounds[name][1] - bounds[name][0])
            elif model.get(name) != 0.:
                step = 0.1 * model.get(name)
            else:
                step = 1.
            start_errors[name] = step

        # count degrees of freedom
        ndof = 0
        if data is not None:
            ndof += len(fitdata)
        elif spectra is not None:
            for spectrum in np.atleast_1d(spectra):
                ndof += len(spectrum)
        ndof -= len(vparam_names)

        # run once with no model covariance, regardless of whether
        # modelcov=True
        fitchisq = generate_chisq(fitdata, model, spectra, signature='iminuit',
                                  modelcov=False)

        m, fmin = _run_iminuit(fitchisq, model.param_names, start_values,
                               start_errors, bounds, fixed_parameters, maxcall,
                               verbose)

        if verbose:
            print("{} function calls; {} dof.".format(fmin.nfcn, ndof))

        # numpy array of best-fit values (including fixed parameters).
        parameters = np.array([m.values[name] for name in model.param_names])
        model.parameters = parameters  # set model parameters to best fit.

        # Iterative Fitting

        if phase_range or wave_range:
            if spectra is not None:
                raise ValueError('phase_range and wave_range are not '
                                 'supported for spectra')
            range_mask = _phase_and_wave_mask(data, model.get('t0'),
                                              model.get('z'),
                                              phase_range, wave_range)
            data_mask = range_mask & support_mask

        # if model covariance, we need to re-run iteratively until convergence
        # if phase range is given, we need to rerun if there are any
        # masked points.
        refit = (modelcov or ((phase_range or wave_range) and
                              np.any(data_mask != support_mask)))

        nfit = 1
        while refit:
            # set new starting point to last point
            for name in vparam_names:
                start_values[name] = model.get(name)

            # re-crop data based on ranges, if necessary
            if (phase_range or wave_range):
                fitdata = data[data_mask]

            # count degrees of freedom
            ndof = 0
            if data is not None:
                ndof += len(fitdata)
            elif spectra is not None:
                for spectrum in np.atleast_1d(spectra):
                    ndof += len(spectrum)
            ndof -= len(vparam_names)

            # generate chisq function based on new starting point
            fitchisq = generate_chisq(fitdata, model, spectra,
                                      signature='iminuit', modelcov=modelcov)

            m, fmin = _run_iminuit(fitchisq, model.param_names, start_values,
                                   start_errors, bounds, fixed_parameters,
                                   maxcall, verbose)

            if verbose:
                print("{} function calls; {} dof.".format(fmin.nfcn, ndof))

            parameters = np.array([m.values[name]
                                   for name in model.param_names])
            model.parameters = parameters
            nfit += 1

            refit = False
            # only consider refitting if we got a valid answer and we're under
            # the maximum number of iterations:
            if fmin.is_valid and nfit < 22:
                # recalculate data mask based on new t0, z
                if phase_range or wave_range:
                    old_data_mask = data_mask
                    range_mask = _phase_and_wave_mask(data, model.get('t0'),
                                                      model.get('z'),
                                                      phase_range, wave_range)
                    data_mask = support_mask & range_mask

                    # we'll refit if we changed any data
                    refit = np.any(data_mask != old_data_mask)

                # refit if *any* parameter changed by more than 10% of
                # statistical error bar
                if modelcov:
                    for name in vparam_names:
                        frac_change = (abs(m.values[name] - start_values[name])
                                       / m.errors[name])
                        refit = refit or frac_change > 0.1

        # Build a message.
        message = []
        if fmin.has_reached_call_limit:
            message.append('Reached call limit.')
        if fmin.hesse_failed:
            message.append('Hesse Failed.')
        if not fmin.has_covariance:
            message.append('No covariance.')
        elif not fmin.has_accurate_covar:  # iminuit docs wrong
            message.append('Covariance may not be accurate.')
        if not fmin.has_posdef_covar:  # iminuit docs wrong
            message.append('Covariance not positive definite.')
        if fmin.has_made_posdef_covar:
            message.append('Covariance forced positive definite.')
        if not fmin.has_valid_parameters:
            message.append('Parameter(s) value and/or error invalid.')
        if len(message) == 0:
            message.append('Minimization exited successfully.')
        # iminuit: m.np_matrix() doesn't work

        # Covariance matrix (only varied parameters) as numpy array.
        if m.covariance is None:
            covariance = None
        else:
            covariance = np.array([
                [m.covariance[(n1, n2)] for n1 in vparam_names]
                for n2 in vparam_names])

        # OrderedDict of errors
        if m.errors is None:
            errors = None
        else:
            errors = OrderedDict((name, m.errors[name])
                                 for name in vparam_names)

        # If we need to, unsort the mask so mask applies to input data
        if sortidx is not None:
            unsort_idx = np.argsort(sortidx)  # indicies that will unsort array
            data_mask = data_mask[unsort_idx]

        # Compile results
        res = Result(success=fmin.is_valid,
                     message=' '.join(message),
                     ncall=fmin.nfcn,
                     chisq=fmin.fval,
                     ndof=ndof,
                     param_names=model.param_names,
                     parameters=parameters,
                     vparam_names=vparam_names,
                     covariance=covariance,
                     errors=errors,
                     nfit=nfit,
                     data_mask=data_mask)

    else:
        raise ValueError("unknown method {0:r}".format(method))

    return res, model


def nest_lc(data, model, vparam_names, bounds, guess_amplitude_bound=False,
            minsnr=5., priors=None, ppfs=None, npoints=100, method='single',
            maxiter=None, maxcall=None, modelcov=False, rstate=None,
            verbose=False, warn=True, **kwargs):
    """Run nested sampling algorithm to estimate model parameters and evidence.

    Parameters
    ----------
    data : `~astropy.table.Table` or `~numpy.ndarray` or `dict`
        Table of photometric data. Must include certain columns.
        See the "Photometric Data" section of the documentation for
        required columns.
    model : `~sncosmo.Model`
        The model to fit.
    vparam_names : list
        Model parameters to vary in the fit.
    bounds : `dict`
        Bounded range for each parameter. Bounds must be given for
        each parameter, with the exception of ``t0``: by default, the
        minimum bound is such that the latest phase of the model lines
        up with the earliest data point and the maximum bound is such
        that the earliest phase of the model lines up with the latest
        data point.
    guess_amplitude_bound : bool, optional
        If true, bounds for the model's amplitude parameter are determined
        automatically based on the data and do not need to be included in
        `bounds`. The lower limit is set to zero and the upper limit is 10
        times the amplitude "guess" (which is based on the highest-flux
        data point in any band). Default is False.
    minsnr : float, optional
        Minimum signal-to-noise ratio of data points to use when guessing
        amplitude bound. Default is 5.
    priors : `dict`, optional
        Prior probability distribution function for each parameter. The keys
        should be parameter names and the values should be callables that
        accept a float. If a parameter is not in the dictionary, the prior
        defaults to a flat distribution between the bounds.
    ppfs : `dict`, optional
        Prior percent point function (inverse of the cumulative distribution
        function) for each parameter. If a parameter is in this dictionary,
        the ppf takes precedence over a prior pdf specified in ``priors``.
    npoints : int, optional
        Number of active samples to use. Increasing this value increases
        the accuracy (due to denser sampling) and also the time
        to solution.
    method : {'classic', 'single', 'multi'}, optional
        Method used to select new points. Choices are 'classic',
        single-ellipsoidal ('single'), multi-ellipsoidal ('multi'). Default
        is 'single'.
    maxiter : int, optional
        Maximum number of iterations. Iteration may stop earlier if
        termination condition is reached. Default is no limit.
    maxcall : int, optional
        Maximum number of likelihood evaluations. Iteration may stop earlier
        if termination condition is reached. Default is no limit.
    modelcov : bool, optional
        Include model covariance when calculating chisq. Default is False.
    rstate : `~numpy.random.RandomState`, optional
        RandomState instance. If not given, the global random state of the
        ``numpy.random`` module will be used.
    verbose : bool, optional
        Print running evidence sum on a single line.
    warn : bool, optional
        Issue warning when dropping bands outside the model range. Default is
        True.

        *New in version 1.5.0*

    Returns
    -------
    res : Result
        Attributes are:

        * ``niter``: total number of iterations
        * ``ncall``: total number of likelihood function calls
        * ``time``: time in seconds spent in iteration loop.
        * ``logz``: natural log of the Bayesian evidence Z.
        * ``logzerr``: estimate of uncertainty in logz (due to finite sampling)
        * ``h``: Bayesian information.
        * ``vparam_names``: list of parameter names varied.
        * ``samples``: 2-d `~numpy.ndarray`, shape is (nsamples, nparameters).
          Each row is the parameter values for a single sample. For example,
          ``samples[0, :]`` is the parameter values for the first sample.
        * ``logprior``: 1-d `~numpy.ndarray` (length=nsamples);
          log(prior volume) for each sample.
        * ``logl``: 1-d `~numpy.ndarray` (length=nsamples); log(likelihood)
          for each sample.
        * ``weights``: 1-d `~numpy.ndarray` (length=nsamples);
          Weight corresponding to each sample. The weight is proportional to
          the prior * likelihood for the sample.
        * ``parameters``: 1-d `~numpy.ndarray` of weighted-mean parameter
          values from samples (including fixed parameters). Order corresponds
          to ``model.param_names``.
        * ``covariance``: 2-d `~numpy.ndarray` of parameter covariance;
          indicies correspond to order of ``vparam_names``. Calculated from
          ``samples`` and ``weights``.
        * ``errors``: OrderedDict of varied parameter uncertainties.
          Corresponds to square root of diagonal entries in covariance matrix.
        * ``ndof``: Number of degrees of freedom (len(data) -
          len(vparam_names)).
        * ``bounds``: Dictionary of bounds on varied parameters (including
          any automatically determined bounds).
        * ``data_mask``: Boolean array the same length as data specifying
          whether each observation was used.
          *New in version 1.5.0.*

    estimated_model : `~sncosmo.Model`
        A copy of the model with parameters set to the values in
        ``res.parameters``.
    """

    try:
        import nestle
    except ImportError:
        raise ImportError("nest_lc() requires the nestle package.")

    # experimental parameters
    tied = kwargs.get("tied", None)

    data = photometric_data(data)

    # sort by time
    if not np.all(np.ediff1d(data.time) >= 0.0):
        sortidx = np.argsort(data.time)
        data = data[sortidx]
    else:
        sortidx = None

    model = copy.copy(model)
    bounds = copy.copy(bounds)  # need to copy this b/c we modify it below

    # Order vparam_names the same way it is ordered in the model:
    vparam_names = [s for s in model.param_names if s in vparam_names]

    # Drop data that the model doesn't cover.
    fitdata, data_mask = cut_bands(data, model,
                                   z_bounds=bounds.get('z', None),
                                   warn=warn)

    if guess_amplitude_bound:
        if model.param_names[2] not in vparam_names:
            raise ValueError("Amplitude bounds guessing enabled but "
                             "amplitude parameter {0!r} is not varied"
                             .format(model.param_names[2]))
        if model.param_names[2] in bounds:
            raise ValueError("cannot supply bounds for parameter {0!r}"
                             " when guess_amplitude_bound=True"
                             .format(model.param_names[2]))

        # If redshift is bounded, set model redshift to midpoint of bounds
        # when doing the guess.
        if 'z' in bounds:
            model.set(z=sum(bounds['z']) / 2.)
        _, amplitude = guess_t0_and_amplitude(fitdata, model, minsnr)
        bounds[model.param_names[2]] = (0., 10. * amplitude)

    # Find t0 bounds to use, if not explicitly given
    if 't0' in vparam_names and 't0' not in bounds:
        bounds['t0'] = t0_bounds(fitdata, model)

    if ppfs is None:
        ppfs = {}
    if tied is None:
        tied = {}

    # Convert bounds/priors combinations into ppfs
    if bounds is not None:
        for key, val in bounds.items():
            if key in ppfs:
                continue  # ppfs take priority over bounds/priors
            a, b = val
            if priors is not None and key in priors:
                # solve ppf at discrete points and return interpolating
                # function
                x_samples = np.linspace(0., 1., 101)
                ppf_samples = ppf(priors[key], x_samples, a, b)
                f = Interp1D(0., 1., ppf_samples)
            else:
                f = Interp1D(0., 1., np.array([a, b]))
            ppfs[key] = f

    # NOTE: It is important that iparam_names is in the same order
    # every time, otherwise results will not be reproducible, even
    # with same random seed.  This is because iparam_names[i] is
    # matched to u[i] below and u will be in a reproducible order,
    # so iparam_names must also be.
    iparam_names = [key for key in vparam_names if key in ppfs]
    ppflist = [ppfs[key] for key in iparam_names]
    npdim = len(iparam_names)  # length of u
    ndim = len(vparam_names)  # length of v

    # Check that all param_names either have a direct prior or are tied.
    for name in vparam_names:
        if name in iparam_names:
            continue
        if name in tied:
            continue
        raise ValueError("Must supply ppf or bounds or tied for parameter '{}'"
                         .format(name))

    def prior_transform(u):
        d = {}
        for i in range(npdim):
            d[iparam_names[i]] = ppflist[i](u[i])
        v = np.empty(ndim, dtype=float)
        for i in range(ndim):
            key = vparam_names[i]
            if key in d:
                v[i] = d[key]
            else:
                v[i] = tied[key](d)
        return v

    # Indicies of the model parameters in vparam_names
    idx = np.array([model.param_names.index(name) for name in vparam_names])

    def loglike(parameters):
        model.parameters[idx] = parameters
        return -0.5 * chisq(fitdata, model, modelcov=modelcov)

    t0 = time.time()
    res = nestle.sample(loglike, prior_transform, ndim, npdim=npdim,
                        npoints=npoints, method=method, maxiter=maxiter,
                        maxcall=maxcall, rstate=rstate,
                        callback=(nestle.print_progress if verbose else None))
    elapsed = time.time() - t0

    # estimate parameters and covariance from samples
    vparameters, cov = nestle.mean_and_cov(res.samples, res.weights)

    # update model parameters to estimated ones.
    model.set(**dict(zip(vparam_names, vparameters)))

    # If we need to, unsort the mask so mask applies to input data
    if sortidx is not None:
        unsort_idx = np.argsort(sortidx)  # indicies that will unsort array
        data_mask = data_mask[unsort_idx]

    # `res` is a nestle.Result object. Collect result into a sncosmo.Result
    # object for consistency, and add more fields.
    res = Result(niter=res.niter,
                 ncall=res.ncall,
                 logz=res.logz,
                 logzerr=res.logzerr,
                 h=res.h,
                 samples=res.samples,
                 weights=res.weights,
                 logvol=res.logvol,
                 logl=res.logl,
                 vparam_names=copy.copy(vparam_names),
                 ndof=len(fitdata) - len(vparam_names),
                 bounds=bounds,
                 time=elapsed,
                 parameters=model.parameters.copy(),
                 covariance=cov,
                 errors=OrderedDict(zip(vparam_names,
                                        np.sqrt(np.diagonal(cov)))),
                 param_dict=OrderedDict(zip(model.param_names,
                                            model.parameters)),
                 data_mask=data_mask)

    return res, model


def mcmc_lc(data, model, vparam_names, bounds=None, priors=None,
            guess_amplitude=True, guess_t0=True, guess_z=True,
            minsnr=5., modelcov=False, nwalkers=10, nburn=200,
            nsamples=1000, sampler='ensemble', thin=1, a=2.0,
            warn=True):
    """Run an MCMC chain to get model parameter samples.

    This is a convenience function around `emcee.EnsembleSampler`.
    It defines the likelihood function and makes a heuristic guess
    at a good set of starting points for the walkers. It then runs
    the sampler, starting with a burn-in run.

    If you're not getting good results, you might want to try
    increasing the burn-in, increasing the walkers, or specifying a
    better starting position.  To get a better starting position, you
    could first run `~sncosmo.fit_lc`, then run this function with all
    ``guess_[name]`` keyword arguments set to False, so that the
    current model parameters are used as the starting point.

    Parameters
    ----------
    data : `~astropy.table.Table` or `~numpy.ndarray` or `dict`
        Table of photometric data. Must include certain columns.
        See the "Photometric Data" section of the documentation for
        required columns.
    model : `~sncosmo.Model`
        The model to fit.
    vparam_names : iterable
        Model parameters to vary.
    bounds : `dict`, optional
        Bounded range for each parameter. Keys should be parameter
        names, values are tuples. If a bound is not given for some
        parameter, the parameter is unbounded. The exception is
        ``t0``: by default, the minimum bound is such that the latest
        phase of the model lines up with the earliest data point and
        the maximum bound is such that the earliest phase of the model
        lines up with the latest data point.
    priors : `dict`, optional
        Prior probability functions. Keys are parameter names, values are
        functions that return probability given the parameter value.
        The default prior is a flat distribution.
    guess_amplitude : bool, optional
        Whether or not to guess the amplitude from the data. If false, the
        current model amplitude is taken as the initial value. Only has an
        effect when fitting amplitude. Default is True.
    guess_t0 : bool, optional
        Whether or not to guess t0. Only has an effect when fitting t0.
        Default is True.
    guess_z : bool, optional
        Whether or not to guess z (redshift). Only has an effect when fitting
        redshift. Default is True.
    minsnr : float, optional
        When guessing amplitude and t0, only use data with signal-to-noise
        ratio (flux / fluxerr) greater than this value. Default is 5.
    modelcov : bool, optional
        Include model covariance when calculating chisq. Default is False.
    nwalkers : int, optional
        Number of walkers in the sampler.
    nburn : int, optional
        Number of samples in burn-in phase.
    nsamples : int, optional
        Number of samples in production run.
    sampler: str, optional
        The kind of sampler to use. Currently only 'ensemble' for
        `emcee.EnsembleSampler` is supported.
    thin : int, optional
        Factor by which to thin samples in production run. Output samples
        array will have (nsamples/thin) samples.
    a : float, optional
        Proposal scale parameter passed to the sampler.
    warn : bool, optional
        Issue a warning when dropping bands outside the wavelength range of
        the model. Default is True.

        *New in version 1.5.0*

    Returns
    -------
    res : Result
        Has the following attributes:

        * ``param_names``: All parameter names of model, including fixed.
        * ``parameters``: Model parameters, with varied parameters set to
          mean value in samples.
        * ``vparam_names``: Names of parameters varied. Order of parameters
          matches order of samples.
        * ``samples``: 2-d array with shape ``(N, len(vparam_names))``.
          Order of parameters in each row  matches order in
          ``res.vparam_names``.
        * ``covariance``: 2-d array giving covariance, measured from samples.
          Order corresponds to ``res.vparam_names``.
        * ``errors``: dictionary giving square root of diagonal of covariance
          matrix for varied parameters. Useful for ``plot_lc``.
        * ``mean_acceptance_fraction``: mean acceptance fraction for all
          walkers in the sampler.
        * ``ndof``: Number of degrees of freedom (len(data) -
          len(vparam_names)).
          *New in version 1.5.0.*
        * ``data_mask``: Boolean array the same length as data specifying
          whether each observation was used.
          *New in version 1.5.0.*

    est_model : `~sncosmo.Model`
        Copy of input model with varied parameters set to mean value in
        samples.

    """

    try:
        import emcee
    except ImportError:
        raise ImportError("mcmc_lc() requires the emcee package.")

    # Standardize and normalize data.
    data = photometric_data(data)

    # sort by time
    if not np.all(np.ediff1d(data.time) >= 0.0):
        sortidx = np.argsort(data.time)
        data = data[sortidx]
    else:
        sortidx = None

    # Make a copy of the model so we can modify it with impunity.
    model = copy.copy(model)

    bounds = copy.deepcopy(bounds) if bounds else {}
    if priors is None:
        priors = {}

    # Check that vparam_names isn't empty, check for unknown parameters.
    if len(vparam_names) == 0:
        raise ValueError("no parameters supplied")
    for names in (vparam_names, bounds, priors):
        for name in names:
            if name not in model.param_names:
                raise ValueError("Parameter not in model: " + repr(name))

    # Order vparam_names the same way it is ordered in the model:
    vparam_names = [s for s in model.param_names if s in vparam_names]
    ndim = len(vparam_names)

    # Check that 'z' is bounded (if it is going to be fit).
    if 'z' in vparam_names:
        if 'z' not in bounds or None in bounds['z']:
            raise ValueError('z must be bounded if allowed to vary.')
        if guess_z:
            model.set(z=sum(bounds['z']) / 2.)
        if model.get('z') < bounds['z'][0] or model.get('z') > bounds['z'][1]:
            raise ValueError('z out of range.')

    # Cut bands that are not allowed by the wavelength range of the model.
    fitdata, data_mask = cut_bands(data, model,
                                   z_bounds=bounds.get('z', None),
                                   warn=warn)

    # Find t0 bounds to use, if not explicitly given
    if 't0' in vparam_names and 't0' not in bounds:
        bounds['t0'] = t0_bounds(fitdata, model)

    # Note that in the parameter guessing below, we assume that the source
    # amplitude is the 3rd parameter of the Model (1st parameter of the Source)

    # Turn off guessing if we're not fitting the parameter.
    if model.param_names[2] not in vparam_names:
        guess_amplitude = False
    if 't0' not in vparam_names:
        guess_t0 = False

    # Make guesses for t0 and amplitude.
    # (we assume amplitude is the 3rd parameter of the model.)
    if guess_amplitude or guess_t0:
        t0, amplitude = guess_t0_and_amplitude(fitdata, model, minsnr)
        if guess_amplitude:
            model.parameters[2] = amplitude
        if guess_t0:
            model.set(t0=t0)

    # Indicies used in probability function.
    # modelidx: Indicies of model parameters corresponding to vparam_names.
    # idxbounds: tuples of (varied parameter index, low bound, high bound).
    # idxpriors: tuples of (varied parameter index, function).
    modelidx = np.array([model.param_names.index(k) for k in vparam_names])
    idxbounds = [(vparam_names.index(k), bounds[k][0], bounds[k][1])
                 for k in bounds]
    idxpriors = [(vparam_names.index(k), priors[k]) for k in priors]

    # Posterior function.
    def lnlike(parameters):
        for i, low, high in idxbounds:
            if not low < parameters[i] < high:
                return -np.inf

        model.parameters[modelidx] = parameters
        logp = -0.5 * chisq(fitdata, model, modelcov=modelcov)
        return logp

    def lnprior(parameters):
        logp = 0
        for i, func in idxpriors:
            logp += math.log(func(parameters[i]))
        return logp

    def lnprob(parameters):
        return lnprior(parameters) + lnlike(parameters)

    # Moves to use
    moves = emcee.moves.StretchMove(a=a)

    if sampler == 'ensemble':
        # Heuristic determination of walker initial positions: distribute
        # walkers in a symmetric gaussian ball, with heuristically
        # determined scale.
        ctr = model.parameters[modelidx]
        scale = np.ones(ndim)
        for i, name in enumerate(vparam_names):
            if name in bounds:
                scale[i] = 0.0001 * (bounds[name][1] - bounds[name][0])
            elif model.get(name) != 0.:
                scale[i] = 0.01 * model.get(name)
            else:
                scale[i] = 0.1
        pos = ctr + scale * np.random.normal(size=(nwalkers, ndim))
        sampler = emcee.EnsembleSampler(nwalkers, ndim, lnprob, moves=moves)

    else:
        raise ValueError('Invalid sampler type. Currently only '
                         '"ensemble" is supported.')

    # Run the sampler.
    pos, prob, state = sampler.run_mcmc(pos, nburn)  # burn-in
    sampler.reset()
    sampler.run_mcmc(pos, nsamples, thin_by=thin)  # production run
    samples = sampler.get_chain(flat=True).reshape(-1, ndim)

    # Summary statistics.
    vparameters = np.mean(samples, axis=0)
    cov = np.cov(samples, rowvar=0)
    model.set(**dict(zip(vparam_names, vparameters)))
    errors = OrderedDict(zip(vparam_names, np.sqrt(np.diagonal(cov))))
    mean_acceptance_fraction = np.mean(sampler.acceptance_fraction)

    # If we need to, unsort the mask so mask applies to input data
    if sortidx is not None:
        unsort_idx = np.argsort(sortidx)  # indicies that will unsort array
        data_mask = data_mask[unsort_idx]

    res = Result(param_names=copy.copy(model.param_names),
                 parameters=model.parameters.copy(),
                 vparam_names=vparam_names,
                 samples=samples,
                 covariance=cov,
                 errors=errors,
                 ndof=len(fitdata) - len(vparam_names),
                 mean_acceptance_fraction=mean_acceptance_fraction,
                 data_mask=data_mask)

    return res, model



# cython: boundscheck=False, wraparound=False, initializedcheck=False
# cython: cdivision=True, auto_pickle=False, language_level=3
# (need auto_pickle=False because we've implemented explicit pickle support via getnewargs() below)
"""
mimic Grid2DFunction function in salt2 software snfit because it doesn't
use spline interpolation; it does bicubic convolution.
"""


import numpy as np
cimport numpy as np
from cpython.mem cimport PyMem_Malloc, PyMem_Free
from libc.math cimport fabs
from libc.string cimport memcpy


cdef int find_index_binary(double *values, int n, double x):
    """Find index i in array such that values[i] <= x < values[i+1].
    using binary search.

    Guaranteed to return values between 0 and n-2 inclusive.
    """
    
    cdef int lo, hi, mid

    lo = 0
    hi = n
    mid = n/2

    if (x < values[0]):
        return 0
    if (x >= values[n-1]):
        return n-2

    while (hi - lo > 1):
        if (x >= values[mid]):
            lo = mid
        else:
            hi = mid
        mid = lo + (hi - lo) / 2

    return mid



cdef int find_index_unsafe(double *values, int n, double x, int start):
    """Return i such that values[i] <= x < values[i+1] via linear search,
    starting from guess `start`.

    If x == values[n-1], n-2 is returned instead of n-1.

    This *assumes* that values[0] <= x <= values[n-1], and that
    0 <= start <= n-2. If that's not true, this will segfault.
    """
    cdef int i

    # search up
    if (x >= values[start]):
        i = start + 1
        while (i < n and x >= values[i]):
            i += 1
        if i == n:
            i -= 1
        return i-1

    # search down
    else:
        i = start - 1
        while (i > -1 and x < values[i]):
            i -= 1
        return i  # -1 should never be returned b/c we assume x >= values[0]


cdef bint is_strictly_ordered(double[:] x):
    cdef int i
    for i in range(1, x.shape[0]):
        if x[i] <= x[i-1]:
            return 0
    return 1


DEF A = -0.5
DEF B = A + 2.0
DEF C = A + 3.0


cdef double kernval(double xval):
     cdef double x = fabs(xval)
     if x > 2.0:
         return 0.0
     if x < 1.0:
         return x * x * (B * x - C) + 1.0
     return A * (-4.0 + x * (8.0 + x * (-5.0 + x)))


cdef class BicubicInterpolator(object):
    """Equivalent of Grid2DFunction in snfit software.

    Bicubic convolution using kernel (and bilinear when next to edge). The
    kernel is defined by:

    x = fabs((distance of point to node)/(distance between nodes)) 

    W(x) = (a+2)*x**3-(a+3)*x**2+1 for x<=1
    W(x) = a( x**3-5*x**2+8*x-4) for 1<x<2
    W(x) = 0 for x>2
    """

    cdef double* xval
    cdef double* yval
    cdef double* fval_storage
    cdef double** fval
    cdef double xmin
    cdef double xmax
    cdef double ymin
    cdef double ymax
    cdef int nx
    cdef int ny

    def __cinit__(self, x, y, z):
        cdef:
            double[:] xc = np.asarray(x, dtype=np.float64)
            double[:] yc = np.asarray(y, dtype=np.float64)
            double[:,:] zc = np.asarray(z, dtype=np.float64)
            int i
            int j

        if not (is_strictly_ordered(xc) and is_strictly_ordered(yc)):
            raise ValueError("x and y values must be strictly increasing")

        self.nx = xc.shape[0]
        self.ny = yc.shape[0]

        # allocate xval
        self.xval = <double *>PyMem_Malloc(self.nx * sizeof(double))
        if not self.xval:
            raise MemoryError()
        for i in range(self.nx):
            self.xval[i] = xc[i]

        self.xmin = self.xval[0]
        self.xmax = self.xval[self.nx-1]

        # allocate yval
        self.yval = <double *>PyMem_Malloc(self.ny * sizeof(double))
        if not self.yval:
            raise MemoryError()
        for i in range(self.ny):
            self.yval[i] = yc[i]

        self.ymin = self.yval[0]
        self.ymax = self.yval[self.ny - 1]

        # copy values array
        self.fval_storage = <double *>PyMem_Malloc(self.nx * self.ny *
                                                   sizeof(double))
        if not self.fval_storage:
            raise MemoryError()
        for i in range(self.nx):
            for j in range(self.ny):
                self.fval_storage[i * self.ny + j] = zc[i, j]

        # allocate fval: pointers to rows of main array
        self.fval = <double **>PyMem_Malloc(self.nx * sizeof(double*))
        if not self.fval:
            raise MemoryError()
        for i in range(self.nx):
            self.fval[i] = self.fval_storage + i * self.ny
        
    def __dealloc__(self):
        PyMem_Free(self.xval)
        PyMem_Free(self.yval)
        PyMem_Free(self.fval)
        PyMem_Free(self.fval_storage)

    def __call__(self, x, y):
        cdef:
            int i,j
            double[:, :] result_view
            double[:] xc = np.atleast_1d(np.asarray(x, dtype=np.float64))
            double[:] yc = np.atleast_1d(np.asarray(y, dtype=np.float64))
            double x_i, y_j
            int ix = 0
            int iy = 0
            double ax, ay, ay2, dx, dy
            int nxc = xc.shape[0]
            int nyc = yc.shape[0]
            double *wyvec
            int *iyvec
            int *yflagvec
            int xflag
            double *wx = [0., 0., 0., 0.]

        # allocate result
        result = np.empty((nxc, nyc), dtype=np.float64)
        result_view = result

        # allocate arrays of y indicies and weights
        # (could use static storage here for small vectors)
        wyvec = <double *>PyMem_Malloc(nyc * 4 * sizeof(double))
        iyvec = <int *>PyMem_Malloc(nyc * sizeof(int))

        # flags: -1 == "skip, return 0", 0 == "linear", 1 == "cubic"
        yflagvec = <int *>PyMem_Malloc(nyc * sizeof(int))

        # find initial indicies by binary search, because they could be
        # anywhere.
        if nxc > 0:
            ix = find_index_binary(self.xval, self.nx, xc[0])
        if nyc > 0:
            iy = find_index_binary(self.yval, self.ny, yc[0])

        # fill above three arrays with y value info
        for j in range(nyc):
            y_j = yc[j]

            # if y is out of range, we won't be using the value at all
            # the inverted comparison here also catches y_j=nan, which
            # would otherwise cause find_index_unsafe() to return -1
            # below, leading to a segfault.
            if (not (y_j >= self.ymin and y_j <= self.ymax)):
                yflagvec[j] = -1
            else:
                iy = find_index_unsafe(self.yval, self.ny, y_j, iy)
                iyvec[j] = iy

                # if we're too close to border, we will use linear
                # interpolation
                # so don't compute weights here
                if (self.ny < 3 or iy == 0 or iy > (self.ny - 3)):
                    yflagvec[j] = 0
                else:
                    # OK to use full cubic interpolation
                    yflagvec[j] = 1

                    # precompute weights
                    dy = ((self.yval[iy] - y_j) /
                          (self.yval[iy+1] - self.yval[iy]))
                    wyvec[4*j+0] = kernval(dy-1.0)
                    wyvec[4*j+1] = kernval(dy)
                    wyvec[4*j+2] = kernval(dy+1.0)
                    wyvec[4*j+3] = kernval(dy+2.0)

        # main loop
        for i in range(nxc):
            x_i = xc[i]

            # precompute some stuff for x
            # again, the inverted comparison shields find_index_unsafe()
            # from invalid input
            if (not (x_i >= self.xmin and x_i <= self.xmax)):
                xflag = -1
            else:
                ix = find_index_unsafe(self.xval, self.nx, x_i, ix)
                if (self.nx < 3 or ix == 0 or ix > (self.nx - 3)):
                    xflag = 0
                else:
                    # OK to use full cubic interpolation
                    xflag = 1

                    # compute weights
                    dx = ((self.xval[ix] - x_i) /
                          (self.xval[ix+1] - self.xval[ix]))
                    wx[0] = kernval(dx-1.0)
                    wx[1] = kernval(dx)
                    wx[2] = kernval(dx+1.0)
                    wx[3] = kernval(dx+2.0)

            # innermost loop
            for j in range(nyc):
                yflag = yflagvec[j]

                # out-of-bounds: return 0.
                if xflag == -1 or yflag == -1:
                    result_view[i, j] = 0.0

                else:
                    iy = iyvec[j]
                    y_j = yc[j]
                    
                    # linear interpolation in *both* dimensions if *either* is
                    # too close to the border. This is how the original code
                    # works, so we mimic it here, even though its dumb.
                    if xflag == 0 or yflag == 0:

                        # If either dimension is 1, just return a close-ish
                        # value
                        if self.nx == 1 or self.ny == 1:
                            result_view[i, j] = self.fval[ix][iy]
                        else:
                            ax = ((x_i - self.xval[ix]) /
                                  (self.xval[ix+1] - self.xval[ix]))
                            ay = ((y_j - self.yval[iy]) /
                                  (self.yval[iy+1] - self.yval[iy]))
                            ay2 = 1.0 - ay
                            
                            result_view[i, j] = (
                                (1.0 - ax) * (ay2 * self.fval[ix  ][iy  ] +
                                              ay  * self.fval[ix  ][iy+1]) +
                                ax         * (ay2 * self.fval[ix+1][iy  ] +
                                              ay  * self.fval[ix+1][iy+1]))

                    # Full cubic convolution
                    else:
                        result_view[i, j] = (
                            wx[0] * (wyvec[4*j+0] * self.fval[ix-1][iy-1] +
                                     wyvec[4*j+1] * self.fval[ix-1][iy  ] +
                                     wyvec[4*j+2] * self.fval[ix-1][iy+1] +
                                     wyvec[4*j+3] * self.fval[ix-1][iy+2]) +
                            wx[1] * (wyvec[4*j+0] * self.fval[ix  ][iy-1] +
                                     wyvec[4*j+1] * self.fval[ix  ][iy  ] +
                                     wyvec[4*j+2] * self.fval[ix  ][iy+1] +
                                     wyvec[4*j+3] * self.fval[ix  ][iy+2]) +
                            wx[2] * (wyvec[4*j+0] * self.fval[ix+1][iy-1] +
                                     wyvec[4*j+1] * self.fval[ix+1][iy  ] +
                                     wyvec[4*j+2] * self.fval[ix+1][iy+1] +
                                     wyvec[4*j+3] * self.fval[ix+1][iy+2]) +
                            wx[3] * (wyvec[4*j+0] * self.fval[ix+2][iy-1] +
                                     wyvec[4*j+1] * self.fval[ix+2][iy  ] +
                                     wyvec[4*j+2] * self.fval[ix+2][iy+1] +
                                     wyvec[4*j+3] * self.fval[ix+2][iy+2]))

        PyMem_Free(iyvec)
        PyMem_Free(wyvec)
        PyMem_Free(yflagvec)

        return result

    def __getnewargs__(self):
        """Return arguments to pass to constructor (to support pickling)"""

        cdef:
            np.ndarray[np.double_t, ndim=1] x
            np.ndarray[np.double_t, ndim=1] y
            np.ndarray[np.double_t, ndim=2] z

        x = np.empty(self.nx, dtype=np.float64)
        y = np.empty(self.ny, dtype=np.float64)
        z = np.empty((self.nx, self.ny), dtype=np.float64)

        memcpy(&x[0], self.xval, self.nx * sizeof(double))
        memcpy(&y[0], self.yval, self.ny * sizeof(double))
        memcpy(&z[0,0], self.fval_storage, self.nx * self.ny * sizeof(double))

        return x, y, z


cdef double polyval(double *coeffs, int n, double x):
    "coeffs[0]*x + coeffs[1]*x^2 + ... + coeffs[n-1]*x^ncoeffs"""

    cdef double out = 0.0
    while n > 0:
        n -= 1
        out = x * (coeffs[n] + out)

    return out


# constants used in SALT2ColorLaw
DEF SALT2CL_B = 4302.57  # B-band-ish wavelength
DEF SALT2CL_V = 5428.55  # V-band-ish wavelength
DEF SALT2CL_V_MINUS_B = SALT2CL_V - SALT2CL_B


cdef class SALT2ColorLaw(object):
    """Callable returning extinction in magnitudes for c=1.

    This is the version 1 extinction law used in SALT2 2.0 (SALT2-2-0)
    and later.

    Parameters
    ----------
    wave_range : (float, float)
    coeffs : list_like


    Notes
    -----
    From snfit code comments:

    if(l_B<=l<=l_R):
        ext = exp(color * constant *
                  (alpha*l + params(0)*l^2 + params(1)*l^3 + ... ))
            = exp(color * constant * P(l))

        where alpha = 1 - params(0) - params(1) - ...

    if (l > l_R):
        ext = exp(color * constant * (P(l_R) + P'(l_R) * (l-l_R)))
    if (l < l_B):
        ext = exp(color * constant * (P(l_B) + P'(l_B) * (l-l_B)))
    """

    cdef:
        int ncoeffs
        double coeffs[7]  # can store up to 6 coeffs (should be only 4)
        double l_lo
        double l_hi
        double p_lo
        double p_hi
        double pprime_lo
        double pprime_hi

    def __cinit__(self, wave_range, coeffs):
        cdef:
            int i
            double wave_lo
            double wave_hi
            double[:] ccoeffs = np.asarray(coeffs, dtype=np.float64)
            double dcoeffs[6]

        if ccoeffs.shape[0] > 6:
            raise ValueError("number of coefficients must be equal to or "
                             "less than 6.")

        # set wave_range
        wave_lo, wave_hi = wave_range
        self.l_lo = (wave_lo - SALT2CL_B) / SALT2CL_V_MINUS_B
        self.l_hi = (wave_hi - SALT2CL_B) / SALT2CL_V_MINUS_B

        for i in range(ccoeffs.shape[0]):
            self.coeffs[i+1] = ccoeffs[i]

        # first coefficient is 'alpha' = 1.0 - sum(other coeffs)
        self.ncoeffs = ccoeffs.shape[0] + 1
        self.coeffs[0] = 1.0
        for i in range(1, self.ncoeffs):
            self.coeffs[0] -= self.coeffs[i]

        # precompute value of
        # P(l) = c[0]*l + c[1]*l^2 + c[2]*l^3 + ...  at l_lo and l_hi
        self.p_lo = polyval(self.coeffs, self.ncoeffs, self.l_lo)
        self.p_hi = polyval(self.coeffs, self.ncoeffs, self.l_hi)

        # precompute derivative of P(l) at l_lo and l_hi
        # P'(l) = c[0] + 2*c[1]*l + 3*c[2]*l^2 + ...)
        for i in range(self.ncoeffs-1):
            dcoeffs[i] = (i+2) * self.coeffs[i+1]  # [2*c[1], 3*c[2], ...]
        self.pprime_lo = self.coeffs[0] + polyval(dcoeffs, self.ncoeffs-1,
                                                  self.l_lo)
        self.pprime_hi = self.coeffs[0] + polyval(dcoeffs, self.ncoeffs-1,
                                                  self.l_hi)

    def __call__(self, double[:] wave):
        cdef:
            double l
            int i, n
            np.ndarray[np.float64_t, ndim=1] out

        n = wave.shape[0]
        out = np.empty(n, dtype=np.float64)

        for i in range(n):
            l = (wave[i] - SALT2CL_B) / SALT2CL_V_MINUS_B

            # Blue side
            if l < self.l_lo:
                out[i] = self.p_lo + self.pprime_lo * (l - self.l_lo)

            # in between
            elif l <= self.l_hi:
                out[i] = polyval(self.coeffs, self.ncoeffs, l)

            # red side
            else:
                out[i] = self.p_hi + self.pprime_hi * (l - self.l_hi)

            out[i] = -out[i]

        return out

    def __getnewargs__(self):
        """Return arguments to pass to constructor (to support pickling)."""

        # Note: an alternative to this would be moving the current __cinit__
        # contents to __init__, then defining __getstate__ and __setstate__
        # to copy the entire struct contents to a bytearray object and back.
        # This would avoid the need to do the initialization again, but
        # seems quite fiddly because the size to copy is unclear:
        # it depends on the struct layout.

        # reconstruct input wavelengths
        wave_lo = self.l_lo * SALT2CL_V_MINUS_B + SALT2CL_B
        wave_hi = self.l_hi * SALT2CL_V_MINUS_B + SALT2CL_B
        coeffs = [self.coeffs[i+1] for i in range(self.ncoeffs)]

        return (wave_lo, wave_hi), coeffs



# Licensed under a 3-clause BSD style license - see LICENSE.rst

"""Convenience functions for interfacing with spectra."""

from scipy.sparse import csr_matrix
import astropy.units as u
import numpy as np

from .bandpasses import get_bandpass
from .constants import HC_ERG_AA, SPECTRUM_BANDFLUX_SPACING, FLAMBDA_UNIT
from .magsystems import get_magsystem
from .utils import integration_grid

__all__ = ['Spectrum']


def _recover_bin_edges(wave):
    """Recover the edges of a set of wavelength bins given the bin centers.

    This function is designed to work for standard linear binning along with
    other more exotic forms of binning such as logarithmic bins. We do a second
    order correction to try to get the bin widths as accurately as possible.
    For linear binning there is only machine precision error with either a
    first or second order estimate.

    For higher order binnings (eg: log), the fractional error is of order (dA /
    A)**2 for linear estimate and (dA / A)**4 for the second order estimate
    that we do here.

    Parameters
    ----------
    wave : array-like
        Central wavelength values of each wavelength bin.

    Returns
    -------
    bin_edges : `~numpy.ndarray`
        The recovered edges of each wavelength bin.
    """
    wave = np.asarray(wave)

    # First order estimate
    o1 = (wave[:-1] + wave[1:]) / 2.

    # Second order correction
    o2 = 1.5*o1[1:-1] - (o1[2:] + o1[:-2]) / 4.

    # Estimate front and back edges
    f2 = 2 * wave[1] - o2[0]
    f1 = 2 * wave[0] - f2
    b2 = 2 * wave[-2] - o2[-1]
    b1 = 2 * wave[-1] - b2

    # Stack everything together
    bin_edges = np.hstack([f1, f2, o2, b2, b1])

    return bin_edges


def _parse_wavelength_information(wave, bin_edges):
    """Parse wavelength information and return a set of bin edges.

    Either the central wavelength for each bin can be passed as ``wave``, or
    the bin edges can be passed directly as ``bin_edges``. This function will
    recover the bin edges from either input and verify that they are a valid
    monotonically-increasing list.
    """
    # Make sure that a valid combination of inputs was given.
    valid_count = 0
    if wave is not None:
        valid_count += 1
    if bin_edges is not None:
        valid_count += 1
    if valid_count != 1:
        raise ValueError('must specify exactly one of wave or bin_edges')

    # Extract the bin starts and ends.
    if wave is not None:
        bin_edges = _recover_bin_edges(np.asarray(wave))

    # Make sure that the bin ends are larger than the bin starts.
    if np.any(bin_edges[1:] <= bin_edges[:-1]):
        raise ValueError('wavelength must be monotonically increasing')

    return bin_edges


class Spectrum(object):
    """An observed spectrum of an object.

    This class is designed to represent an observed spectrum. An observed
    spectrum is a set of contiguous bins in wavelength (referred to as
    "spectral elements") with associated flux measurements. We assume that each
    spectral element has uniform transmission in wavelength. A spectrum can
    optionally have associated uncertainties or covariance between the observed
    fluxes of the different spectral elements. A spectrum can also optionally
    have a time associated with it.

    Internally, we store the edges of each of the spectral element wavelength
    bins. These are automatically recovered in the common case where a user has
    a list of central wavelengths for each bin. The wavelengths are stored
    internally in units of Angstroms. The flux is stored as a spectral flux
    density F_λ (units of erg / s / cm^2 / Angstrom).

    Parameters
    ----------
    wave : list-like
        Central wavelengths of each spectral element. This must be
        monotonically increasing. This is assumed to be in units of Angstroms
        unless ``wave_unit`` is specified.
    flux : list-like
        Observed fluxes for each spectral element. By default this is assumed
        to be a spectral flux density F_λ unless ``unit`` is explicitly
        specified.
    fluxerr : list-like
        Uncertainties on the observed fluxes for each spectral element.
    fluxcov : two-dimensional `~numpy.ndarray`
        Covariance of the observed fluxes for each spectral element. Only one
        of ``fluxerr`` and ``fluxcov`` may be specified.
    bin_edges : list-like
        Edges of each spectral element in wavelength. This should be a list
        that is length one longer than ``flux``. Only one of ``wave`` and
        ``bin_edges`` may be specified.
    wave_unit : `~astropy.units.Unit`
        Wavelength unit. Default is Angstroms.
    unit : `~astropy.units.Unit`
        Flux unit. Default is F_λ (erg / s / cm^2 / Angstrom).
    time : float
        The time associated with the spectrum. This is required if fitting a
        model to the spectrum.
    """
    def __init__(self, wave=None, flux=None, fluxerr=None, fluxcov=None,
                 bin_edges=None, wave_unit=u.AA, unit=FLAMBDA_UNIT, time=None):
        # Extract the bin edges
        bin_edges = _parse_wavelength_information(wave, bin_edges)
        self.bin_edges = bin_edges

        # Make sure that the flux data matches up with the wavelength data.
        self.flux = np.asarray(flux)
        if not (len(self.bin_edges) - 1 == len(self.flux)):
            raise ValueError("unequal column lengths")

        # Extract uncertainty information in whatever form it came in.
        self._fluxerr = None
        self._fluxcov = None

        if fluxerr is not None:
            if fluxcov is not None:
                raise ValueError("can only specify one of fluxerr and fluxcov")
            self._fluxerr = np.array(fluxerr)
            if len(self._fluxerr) != len(self.flux):
                raise ValueError("unequal column lengths")
        elif fluxcov is not None:
            self._fluxcov = np.array(fluxcov)
            if not (len(self.flux) == self._fluxcov.shape[0] ==
                    self._fluxcov.shape[1]):
                raise ValueError("unequal column lengths")

        # Internally, wavelength is in Angstroms:
        if wave_unit != u.AA:
            self.bin_edges = wave_unit.to(u.AA, self.bin_edges, u.spectral())
        self._wave_unit = u.AA

        # Internally, flux is in F_lambda:
        if unit != FLAMBDA_UNIT:
            unit_scale = unit.to(
                FLAMBDA_UNIT, equivalencies=u.spectral_density(u.AA, self.wave)
            )
            self.flux = unit_scale * self.flux
            if self._fluxerr is not None:
                self._fluxerr = unit_scale * self._fluxerr
            if self._fluxcov is not None:
                self._fluxcov = np.outer(unit_scale, unit_scale) \
                    .dot(self._fluxcov)
        self._unit = FLAMBDA_UNIT

        self.time = time

        # We use a sampling matrix to evaluate models/bands for the spectrum.
        # This matrix is expensive to compute but rarely changes, so we cache
        # it.
        self._cache_sampling_matrix = None

    def __len__(self):
        return len(self.flux)

    @property
    def bin_starts(self):
        """Return the start of each bin."""
        return self.bin_edges[:-1]

    @property
    def bin_ends(self):
        """Return the end of each bin."""
        return self.bin_edges[1:]

    @property
    def wave(self):
        """Return the centers of each bin."""
        return (self.bin_starts + self.bin_ends) / 2.

    @property
    def fluxerr(self):
        """Return the uncertainties on each flux bin"""
        if self._fluxerr is not None:
            return self._fluxerr
        elif self._fluxcov is not None:
            return np.sqrt(np.diag(self._fluxcov))
        else:
            raise ValueError("no uncertainty information available")

    @property
    def fluxcov(self):
        """Return the covariance matrix"""
        if self._fluxcov is not None:
            return self._fluxcov
        elif self._fluxerr is not None:
            return np.diag(self._fluxerr**2)
        else:
            raise ValueError("no uncertainty information available")

    def has_uncertainties(self):
        """Check whether there is uncertainty information available."""
        return self._fluxcov is not None or self._fluxerr is not None

    def rebin(self, wave=None, bin_edges=None):
        """Rebin the spectrum on a new wavelength grid.

        We assume that the spectrum is constant for each spectral element with
        a value given by its observed flux. If the new bin edges are not
        aligned with the old ones, then this will introduce covariance between
        spectral elements. We propagate that covariance properly.

        Parameters
        ----------
        wave : list-like
            Central wavelengths of the rebinned spectrum.
        bin_edges : list-like
            Bin edges of the rebinned spectrum. Only one of ``wave`` and
            ``bin_edges`` may be specified.

        Returns
        -------
        rebinned_spectrum : `~sncosmo.Spectrum`
            A new `~sncosmo.Spectrum` with the rebinned spectrum.
        """
        new_bin_edges = _parse_wavelength_information(wave, bin_edges)
        new_bin_starts = new_bin_edges[:-1]
        new_bin_ends = new_bin_edges[1:]

        old_bin_starts = self.bin_starts
        old_bin_ends = self.bin_ends

        # Generate a weight matrix for the transformation.
        overlap_starts = np.max(np.meshgrid(old_bin_starts, new_bin_starts),
                                axis=0)
        overlap_ends = np.min(np.meshgrid(old_bin_ends, new_bin_ends), axis=0)
        overlaps = overlap_ends - overlap_starts
        overlaps[overlaps < 0] = 0

        # Normalize by the total overlap in each bin to keep everything in
        # units of f_lambda
        total_overlaps = np.sum(overlaps, axis=1)
        if np.any(total_overlaps == 0):
            raise ValueError("new binning not contained within original "
                             "spectrum")
        weight_matrix = overlaps / total_overlaps[:, None]

        new_flux = weight_matrix.dot(self.flux)
        new_fluxcov = weight_matrix.dot(self.fluxcov.dot(weight_matrix.T))

        return Spectrum(
            bin_edges=new_bin_edges,
            flux=new_flux,
            fluxcov=new_fluxcov,
            time=self.time,
        )

    def get_sampling_matrix(self):
        """Build an appropriate sampling for the spectral elements.

        For spectra with wide spectral elements, it is important to integrate
        models over the spectral element rather than simply sampling at the
        central wavelength. This function first determines where to sample for
        each spectral element and returns the corresponding list of wavelengths
        ``sample_wave``. This function also returns a matrix
        ``sampling_matrix`` that provided the mapping between the sampled
        wavelengths and the spectral elements. Given a set of model fluxes
        evaluated at ``sample_wave``, the dot product of ``sampling_matrix``
        with these fluxes gives the corresponding fluxes in each spectral
        element in units of (erg / s / cm^2).

        ``sampling_matrix`` is stored as a compressed sparse row matrix that
        can be very efficiently used for dot products with vectors. This matrix
        is somewhat expensive to calculate and only changes if the bin edges of
        the spectral elements change, so we cache it and only recompute it if
        the bin edges change.

        Returns
        -------
        sample_wave : `~numpy.ndarray`
            Wavelengths to sample a model at.
        sampling_matrix : `~scipy.sparse.csr_matrix`
            Matrix giving the mapping from the sampled bins to the spectral
            elements.
        """
        # Check if we have cached the sampling matrix already.
        if self._cache_sampling_matrix is not None:
            cache_bin_edges, sampling_matrix_result = \
                self._cache_sampling_matrix
            if np.all(cache_bin_edges == self.bin_edges):
                # No changes to the spectral elements so the sampling matrix
                # hasn't changed.
                return sampling_matrix_result

        indices = []
        sample_wave = []
        sample_dwave = []

        for bin_idx in range(len(self.flux)):
            bin_start = self.bin_starts[bin_idx]
            bin_end = self.bin_ends[bin_idx]

            bin_wave, bin_dwave = integration_grid(bin_start, bin_end,
                                                   SPECTRUM_BANDFLUX_SPACING)

            indices.append(bin_idx * np.ones_like(bin_wave, dtype=int))
            sample_wave.append(bin_wave)
            sample_dwave.append(bin_dwave * np.ones_like(bin_wave))

        indices = np.hstack(indices)
        sample_wave = np.hstack(sample_wave)
        sample_dwave = np.hstack(sample_dwave)

        sampling_matrix = csr_matrix(
            (sample_dwave, (indices, np.arange(len(indices)))),
            shape=(len(self), len(indices)),
            dtype=np.float64,
        )

        # Cache the result
        sampling_matrix_result = (sample_wave, sampling_matrix)
        self._cache_sampling_matrix = (self.bin_edges.copy(),
                                       sampling_matrix_result)

        return sampling_matrix_result

    def _band_weights(self, band, zp, zpsys):
        """Calculate the weights for synthetic photometry.

        Parameters
        ----------
        band : `~sncosmo.Bandpass`, str or list_like
            Bandpass, name of bandpass in registry, or list or array thereof.
        zp : float or list_like, optional
            If given, zeropoint to scale flux to (must also supply ``zpsys``).
            If not given, flux is not scaled.
        zpsys : str or list_like, optional
            Name of a magnitude system in the registry, specifying the system
            that ``zp`` is in.

        Returns
        -------
        band_weights : numpy.array
            The weights to multiply each bin by for synthetic photometry in the
            given band(s). This has a shape of (number of bands, number of
            spectral elements). The dot product of this array with the flux
            array gives the desired band flux.
        """
        band_weights = []

        if zp is not None and zpsys is None:
            raise ValueError('zpsys must be given if zp is not None')

        # broadcast arrays
        if zp is None:
            band = np.atleast_1d(band)
        else:
            band, zp, zpsys = np.broadcast_arrays(np.atleast_1d(band), zp,
                                                  zpsys)

        for idx in range(len(band)):
            iter_band = get_bandpass(band[idx])

            # Check that bandpass wavelength range is fully contained in
            # spectrum wavelength range.
            if (iter_band.minwave() < self.bin_starts[0]
                    or iter_band.maxwave() > self.bin_ends[-1]):
                raise ValueError(
                    'bandpass {0!r:s} [{1:.6g}, .., {2:.6g}] '
                    'outside spectral range [{3:.6g}, .., {4:.6g}]'
                    .format(iter_band.name, iter_band.minwave(),
                            iter_band.maxwave(), self.bin_starts[0],
                            self.bin_ends[-1])
                )

            sample_wave, sampling_matrix = self.get_sampling_matrix()
            trans = iter_band(sample_wave)
            sample_weights = sample_wave * trans / HC_ERG_AA
            row_band_weights = sampling_matrix * sample_weights

            if zp is not None:
                ms = get_magsystem(zpsys[idx])
                zp_bandflux = ms.zpbandflux(iter_band)
                zpnorm = 10. ** (0.4 * zp[idx]) / zp_bandflux
                row_band_weights *= zpnorm

            band_weights.append(row_band_weights)

        band_weights = np.vstack(band_weights)

        return band_weights

    def bandflux(self, band, zp=None, zpsys=None):
        """Perform synthentic photometry in a given bandpass.

        We assume that the spectrum is constant for each spectral element with
        a value given by its observed flux. The bandpass is sampled on an
        appropriate high-resolution grid and multiplied with the observed
        fluxes to give the corresponding integrated band flux over this band.

        Parameters
        ----------
        band : `~sncosmo.bandpass` or str or list_like
            Bandpass(es) or name(s) of bandpass(es) in registry.
        zp : float or list_like, optional
            If given, zeropoint to scale flux to. if `none` (default) flux
            is not scaled.
        zpsys : `~sncosmo.magsystem` or str (or list_like), optional
            Determines the magnitude system of the requested zeropoint.
            cannot be `none` if `zp` is not `none`.

        Returns
        -------
        bandflux : float or `~numpy.ndarray`
            Flux in photons / s / cm^2, unless `zp` and `zpsys` are
            given, in which case flux is scaled so that it corresponds
            to the requested zeropoint. Return value is `float` if all
            input parameters are scalars, `~numpy.ndarray` otherwise.
        """
        ndim = np.ndim(band)

        band_weights = self._band_weights(band, zp, zpsys)
        band_flux = band_weights.dot(self.flux)

        if ndim == 0:
            band_flux = band_flux[0]

        return band_flux

    def bandfluxcov(self, band, zp=None, zpsys=None):
        """Like bandflux(), but also returns model covariance on values.

        Parameters
        ----------
        band : `~sncosmo.bandpass` or str or list_like
            Bandpass(es) or name(s) of bandpass(es) in registry.
        zp : float or list_like, optional
            If given, zeropoint to scale flux to. if `none` (default) flux
            is not scaled.
        zpsys : `~sncosmo.magsystem` or str (or list_like), optional
            Determines the magnitude system of the requested zeropoint.
            cannot be `none` if `zp` is not `none`.

        Returns
        -------
        bandflux : float or `~numpy.ndarray`
            Model bandfluxes.
        cov : float or `~numpy.array`
            Covariance on ``bandflux``. If ``bandflux`` is an array, this
            will be a 2-d array.
        """
        ndim = np.ndim(band)

        band_weights = self._band_weights(band, zp, zpsys)
        band_flux = band_weights.dot(self.flux)
        band_cov = band_weights.dot(self.fluxcov).dot(band_weights.T)

        if ndim == 0:
            band_flux = band_flux[0]
            band_cov = band_cov[0, 0]

        return band_flux, band_cov

    def bandmag(self, band, magsys):
        """Magnitude through the given bandpass(es), and for the given
        magnitude system(s).

        Parameters
        ----------
        band : str or list_like
            Name(s) of bandpass in registry.
        magsys : str or list_like
            Name(s) of `~sncosmo.MagSystem` in registry.

        Returns
        -------
        mag : float or `~numpy.ndarray`
            Magnitude for each item in band, magsys.
            The return value is a float if all parameters are not iterables.
            The return value is an `~numpy.ndarray` if any are iterable.
        """
        return -2.5 * np.log10(self.bandflux(band, 0., magsys))



import numpy as np
import pytest

from astropy.table import Table
from numpy.testing import assert_allclose

import sncosmo


try:
    import iminuit

    HAS_IMINUIT = True

except ImportError:
    HAS_IMINUIT = False


def test_bin_edges_linear():
    """Ensure that we can recover consistent bin edges for a spectrum from bin
    centers.

    Internally, the bin edges are stored rather than the bin centers.
    """
    wave = np.linspace(3000, 8000, 100)
    flux = np.ones_like(wave)
    spec = sncosmo.Spectrum(wave, flux)

    assert_allclose(wave, spec.wave, rtol=1.e-5)


def test_bin_edges_log():
    """Ensure that we can recover consistent bin edges for a spectrum from bin
    centers.

    Internally, the bin edges are stored rather than the bin centers.
    """
    wave = np.logspace(np.log10(3000), np.log10(8000), 100)
    flux = np.ones_like(wave)
    spec = sncosmo.Spectrum(wave, flux)

    assert_allclose(wave, spec.wave, rtol=1.e-5)


@pytest.mark.might_download
class TestSpectrum:
    def setup_class(self):
        # Simulate a spectrum
        model = sncosmo.Model(source='hsiao-subsampled')
        params = {'t0': 10., 'amplitude': 1.e-7, 'z': 0.2}
        start_params = {'t0': 0., 'amplitude': 1., 'z': 0.}
        model.set(**params)

        # generate a fake spectrum with no errors. note: we simulate a high
        # resolution spectrum and then bin it up. we also include large
        # covariance between spectral elements to verify that we are handling
        # covariance properly.
        spec_time = params['t0'] + 5.
        sim_wave = np.arange(3000, 9000)
        sim_flux = model.flux(spec_time, sim_wave)
        sim_fluxcov = 0.01 * np.max(sim_flux)**2 * np.ones((len(sim_flux),
                                                            len(sim_flux)))
        sim_fluxcov += np.diag(0.1 * sim_flux**2)
        spectrum = sncosmo.Spectrum(sim_wave, sim_flux, fluxcov=sim_fluxcov,
                                    time=spec_time)

        # generate a binned up low-resolution spectrum.
        bin_wave = np.linspace(3500, 8500, 200)
        bin_spectrum = spectrum.rebin(bin_wave)

        # generate fake photometry with no errors
        points_per_band = 12
        bands = points_per_band * ['bessellux', 'bessellb', 'bessellr',
                                   'besselli']
        times = params['t0'] + np.linspace(-10., 60., len(bands))
        zp = len(bands) * [25.]
        zpsys = len(bands) * ['ab']
        flux = model.bandflux(bands, times, zp=zp, zpsys=zpsys)
        fluxerr = len(bands) * [0.1 * np.max(flux)]
        photometry = Table({
            'time': times,
            'band': bands,
            'flux': flux,
            'fluxerr': fluxerr,
            'zp': zp,
            'zpsys': zpsys
        })

        self.model = model
        self.photometry = photometry
        self.spectrum = spectrum
        self.bin_spectrum = bin_spectrum
        self.params = params
        self.start_params = start_params

    def test_bandflux(self):
        """Check synthetic photometry.

        We compare synthetic photometry on high and low resolution spectra. It
        should stay the same.
        """
        bandflux_highres = self.spectrum.bandflux('sdssg')
        bandflux_lowres = self.bin_spectrum.bandflux('sdssg')

        assert_allclose(bandflux_highres, bandflux_lowres, rtol=1.e-3)

    def test_bandflux_multi(self):
        """Check synthetic photometry with multiple bands."""
        bands = ['sdssg', 'sdssr', 'sdssi']
        bandflux_highres = self.spectrum.bandflux(bands)
        bandflux_lowres = self.bin_spectrum.bandflux(bands)

        assert_allclose(bandflux_highres, bandflux_lowres, rtol=1.e-3)

    def test_bandflux_zpsys(self):
        """Check synthetic photometry with a magnitude system."""
        bands = ['sdssg', 'sdssr', 'sdssi']
        bandflux_highres = self.spectrum.bandflux(bands, 25., 'ab')
        bandflux_lowres = self.spectrum.bandflux(bands, 25., 'ab')

        assert_allclose(bandflux_highres, bandflux_lowres, rtol=1.e-3)

    def test_bandfluxcov(self):
        """Check synthetic photometry with covariance."""
        bands = ['sdssg', 'sdssr', 'sdssi']
        flux_highres, cov_highres = self.spectrum.bandfluxcov(bands)
        flux_lowres, cov_lowres = self.bin_spectrum.bandfluxcov(bands)

        assert_allclose(flux_highres, flux_lowres, rtol=1.e-3)
        assert_allclose(cov_highres, cov_lowres, rtol=1.e-3)

    def test_bandmag(self):
        """Check synthetic photometry in magnitudes."""
        bands = ['sdssg', 'sdssr', 'sdssi']
        bandmag_highres = self.spectrum.bandmag(bands, 'ab')
        bandmag_lowres = self.bin_spectrum.bandmag(bands, 'ab')

        assert_allclose(bandmag_highres, bandmag_lowres, rtol=1.e-3)

    @pytest.mark.skipif('not HAS_IMINUIT')
    def test_fit_lc_spectra(self):
        """Check fit results for a single high-resolution spectrum."""
        self.model.set(**self.start_params)
        res, fitmodel = sncosmo.fit_lc(model=self.model,
                                       spectra=self.bin_spectrum,
                                       vparam_names=['amplitude', 'z', 't0'],
                                       bounds={'z': (0., 0.3)})

        # set model to true parameters and compare to fit results.
        self.model.set(**self.params)
        assert_allclose(res.parameters, self.model.parameters, rtol=1.e-3)

    @pytest.mark.skipif('not HAS_IMINUIT')
    def test_fit_lc_both(self):
        """Check fit results for both spectra and photometry."""
        self.model.set(**self.start_params)
        res, fitmodel = sncosmo.fit_lc(self.photometry, model=self.model,
                                       spectra=self.bin_spectrum,
                                       vparam_names=['amplitude', 'z', 't0'],
                                       bounds={'z': (0., 0.3)})

        # set model to true parameters and compare to fit results.
        self.model.set(**self.params)
        assert_allclose(res.parameters, self.model.parameters, rtol=1.e-3)

    @pytest.mark.skipif('not HAS_IMINUIT')
    def test_fit_lc_multiple_spectra(self):
        """Check fit results for multiple spectra."""
        self.model.set(**self.start_params)
        res, fitmodel = sncosmo.fit_lc(model=self.model,
                                       spectra=[self.bin_spectrum,
                                                self.bin_spectrum],
                                       vparam_names=['amplitude', 'z', 't0'],
                                       bounds={'z': (0., 0.3)})

        # set model to true parameters and compare to fit results.
        self.model.set(**self.params)
        assert_allclose(res.parameters, self.model.parameters, rtol=1.e-3)



# Licensed under a 3-clause BSD style license - see LICENSES

from io import StringIO

import scipy

import numpy as np
from numpy.testing import assert_allclose, assert_approx_equal

import sncosmo
from sncosmo import registry


def flatsource():
    """Create and return a TimeSeriesSource with a flat spectrum == 1.0 at
    all times."""
    phase = np.linspace(0., 100., 10)
    wave = np.linspace(800., 20000., 100)
    flux = np.ones((len(phase), len(wave)), dtype=float)
    return sncosmo.TimeSeriesSource(phase, wave, flux)


class AchromaticMicrolensing(sncosmo.PropagationEffect):
    """
    An achromatic microlensing object. Tests phase dependence.
    """
    _param_names = []
    param_names_latex = []
    _minwave = np.nan
    _maxwave = np.nan
    _minphase = np.nan
    _maxphase = np.nan

    def __init__(self, time, magnification):
        """
        Parameters
            ----------
            time: :class:`~list` or :class:`~numpy.array`
                A time array for your microlensing
            dmag: :class:`~list` or :class:`~numpy.array`
                microlensing magnification
        """
        self._parameters = np.array([])
        self.mu = scipy.interpolate.interp1d(time,
                                             magnification,
                                             bounds_error=False,
                                             fill_value=1.)
        self._minphase = np.min(time)
        self._maxphase = np.max(time)

    def propagate(self, wave, flux, phase):
        """
        Propagate the magnification onto the model's flux output.
        """

        mu = np.expand_dims(np.atleast_1d(self.mu(phase)), 1)
        return flux * mu


class StepEffect(sncosmo.PropagationEffect):
    """
    Effect with transmission 0 below cutoff wavelength, 1 above.
    Useful for testing behavior with redshift.
    """

    _param_names = ['stepwave']
    param_names_latex = [r'\lambda_{s}']

    def __init__(self, minwave, maxwave, stepwave=10000.):
        self._minwave = minwave
        self._maxwave = maxwave
        self._parameters = np.array([stepwave], dtype=np.float64)

    def propagate(self, wave, flux):
        return flux * (wave >= self._parameters[0])


class TestTimeSeriesSource:

    def setup_class(self):
        self.source = flatsource()

    def test_flux(self):
        for a in [1., 2.]:
            self.source.set(amplitude=a)
            assert_allclose(self.source.flux(1., 2000.), a)
            assert_allclose(self.source.flux(1., [2000.]), np.array([a]))
            assert_allclose(self.source.flux([1.], 2000.), np.array([[a]]))
            assert_allclose(self.source.flux([1.], [2000.]), np.array([[a]]))

    def test_getsource(self):
        # register the source & retrieve it from the registry
        registry.register(self.source, name="testsource",
                          data_class=sncosmo.Source)
        m = sncosmo.get_source("testsource")

    def test_bandflux(self):
        self.source.set(amplitude=1.0)
        f = self.source.bandflux("bessellb", 0.)

        # Correct answer
        b = sncosmo.get_bandpass("bessellb")
        dwave = np.gradient(b.wave)
        ans = np.sum(b.trans * b.wave * dwave) / sncosmo.models.HC_ERG_AA
        assert_approx_equal(ans, f)

    def test_bandflux_shapes(self):
        # Just check that these work.
        self.source.bandflux("bessellb", 0., zp=25., zpsys="ab")
        self.source.bandflux("bessellb", [0.1, 0.2], zp=25., zpsys="ab")
        self.source.bandflux(["bessellb", "bessellv"], [0.1, 0.2], zp=25.,
                             zpsys="ab")


class TestSALT2Source:
    def setup_class(self):
        """Create a SALT2 model with a lot of components set to 1."""

        phase = np.linspace(0., 100., 10)
        wave = np.linspace(1000., 10000., 100)
        vals1d = np.zeros(len(phase), dtype=np.float64)
        vals = np.ones([len(phase), len(wave)], dtype=np.float64)

        # Create some 2-d grid files
        files = []
        for i in [0, 1]:
            f = StringIO()
            sncosmo.write_griddata_ascii(phase, wave, vals, f)
            f.seek(0)  # return to start of file.
            files.append(f)

        # CL file. The CL in magnitudes will be
        # CL(wave) = -(wave - B) / (V - B)  [B = 4302.57, V = 5428.55]
        # and transmission will be 10^(-0.4 * CL(wave))^c
        clfile = StringIO()
        clfile.write("1\n"
                     "0.0\n"
                     "Salt2ExtinctionLaw.version 1\n"
                     "Salt2ExtinctionLaw.min_lambda 3000\n"
                     "Salt2ExtinctionLaw.max_lambda 7000\n")
        clfile.seek(0)

        # Create some more 2-d grid files
        for factor in [1., 0.01, 0.01, 0.01]:
            f = StringIO()
            sncosmo.write_griddata_ascii(phase, wave, factor * vals, f)
            f.seek(0)  # return to start of file.
            files.append(f)

        # Create a 1-d grid file (color dispersion)
        cdfile = StringIO()
        for w in wave:
            cdfile.write("{0:f} {1:f}\n".format(w, 0.2))
        cdfile.seek(0)  # return to start of file.

        # Create a SALT2Source
        self.source = sncosmo.SALT2Source(m0file=files[0],
                                          m1file=files[1],
                                          clfile=clfile,
                                          errscalefile=files[2],
                                          lcrv00file=files[3],
                                          lcrv11file=files[4],
                                          lcrv01file=files[5],
                                          cdfile=cdfile)

        for f in files:
            f.close()
        cdfile.close()
        clfile.close()

    def test_bandflux_rcov(self):

        # component 1:
        # ans = (F0/F1)^2 S^2 (V00 + 2 x1 V01 + x1^2 V11)
        # when x1=0, this reduces to S^2 V00 = 1^2 * 0.01 = 0.01
        #
        # component 2:
        # cd^2 = 0.04
        phase = np.linspace(0., 100., 10)
        wave = np.linspace(1000., 10000., 100)
        vals = np.ones([len(phase), len(wave)], dtype=np.float64)

        band = ['bessellb', 'bessellb', 'bessellr', 'bessellr',
                'besselli']
        band = np.array([sncosmo.get_bandpass(b) for b in band])
        phase = np.array([10., 20., 30., 40., 50.])
        self.source.set(x1=0.0)
        result = self.source.bandflux_rcov(band, phase)
        expected = np.array([[0.05, 0.04, 0.,   0.,   0.],
                             [0.04, 0.05, 0.,   0.,   0.],
                             [0.,   0.,   0.05, 0.04, 0.],
                             [0.,   0.,   0.04, 0.05, 0.],
                             [0.,   0.,   0.,   0.,   0.05]])
        assert_allclose(result, expected)


class TestSALT3Source:

    def setup_class(self):
        """Create a SALT3 model with a lot of components set to 1."""

        phase = np.linspace(0., 100., 10)
        wave = np.linspace(1000., 10000., 100)
        vals1d = np.zeros(len(phase), dtype=np.float64)
        vals = np.ones([len(phase), len(wave)], dtype=np.float64)

        # Create some 2-d grid files
        files = []
        for i in [0, 1]:
            f = StringIO()
            sncosmo.write_griddata_ascii(phase, wave, vals, f)
            f.seek(0)  # return to start of file.
            files.append(f)

        # CL file. The CL in magnitudes will be
        # CL(wave) = -(wave - B) / (V - B)  [B = 4302.57, V = 5428.55]
        # and transmission will be 10^(-0.4 * CL(wave))^c
        clfile = StringIO()
        clfile.write("1\n"
                     "0.0\n"
                     "Salt2ExtinctionLaw.version 1\n"
                     "Salt2ExtinctionLaw.min_lambda 3000\n"
                     "Salt2ExtinctionLaw.max_lambda 8000\n")
        clfile.seek(0)

        # Create some more 2-d grid files
        for factor in [1., 0.01, 0.01, 0.01]:
            f = StringIO()
            sncosmo.write_griddata_ascii(phase, wave, factor * vals, f)
            f.seek(0)  # return to start of file.
            files.append(f)

        # Create a 1-d grid file (color dispersion)
        cdfile = StringIO()
        for w in wave:
            cdfile.write("{0:f} {1:f}\n".format(w, 0.2))
        cdfile.seek(0)  # return to start of file.

        # Create a SALT2Source
        self.source = sncosmo.SALT3Source(m0file=files[0],
                                          m1file=files[1],
                                          clfile=clfile,
                                          lcrv00file=files[3],
                                          lcrv11file=files[4],
                                          lcrv01file=files[5],
                                          cdfile=cdfile)

        for f in files:
            f.close()
        cdfile.close()
        clfile.close()

    def test_bandflux_rcov(self):

        # component 1:
        # ans = (F0/F1)^2 S^2 (V00 + 2 x1 V01 + x1^2 V11)
        # when x1=0, this reduces to S^2 V00 = 1^2 * 0.01 = 0.01
        #
        # component 2:
        # cd^2 = 0.04
        phase = np.linspace(0., 100., 10)
        wave = np.linspace(1000., 10000., 100)
        vals = np.ones([len(phase), len(wave)], dtype=np.float64)
        band = ['bessellb', 'bessellb', 'bessellr', 'bessellr',
                'besselli']
        band = np.array([sncosmo.get_bandpass(b) for b in band])
        phase = np.array([10., 20., 30., 40., 50.])
        self.source.set(x1=0.0)
        result = self.source.bandflux_rcov(band, phase)
        expected = np.array([[0.05, 0.04, 0.,   0.,   0.],
                             [0.04, 0.05, 0.,   0.,   0.],
                             [0.,   0., 0.05, 0.04, 0.],
                             [0.,   0., 0.04, 0.05, 0.],
                             [0.,   0., 0.,   0.,  0.05]])
        assert_allclose(result, expected)


class TestModel:

    def setup_class(self):
        self.model = sncosmo.Model(source=flatsource(),
                                   effects=[sncosmo.CCM89Dust()],
                                   effect_frames=['obs'],
                                   effect_names=['mw'])

    def test_minwave(self):
        # at redshift zero, should be determined by effect minwave (~909)
        self.model.set(z=0.)
        ans = max(self.model.source.minwave(),
                  self.model.effects[0].minwave())
        assert self.model.minwave() == ans

        # at redshift 1, should be determined by effect minwave (800*2)
        z = 1.
        self.model.set(z=z)
        ans = max(self.model.source.minwave() * (1.+z),
                  self.model.effects[0].minwave())
        assert self.model.minwave() == ans

    def test_maxwave(self):
        # at redshift zero, should be determined by source maxwave (20000)
        self.model.set(z=0.)
        ans = min(self.model.source.maxwave(),
                  self.model.effects[0].maxwave())
        assert self.model.maxwave() == ans

        # at redshift 1, should be determined by effect maxwave (33333)
        z = 1.
        self.model.set(z=z)
        ans = min(self.model.source.maxwave() * (1.+z),
                  self.model.effects[0].maxwave())
        assert self.model.maxwave() == ans

    def test_set_source_peakabsmag(self):
        # Both Bandpass and str should work
        band = sncosmo.get_bandpass('bessellb')
        self.model.set_source_peakabsmag(-19.3, 'bessellb', 'ab')
        self.model.set_source_peakabsmag(-19.3, band, 'ab')

    def test_str(self):
        """Test if string summary works at all."""
        str(self.model)

    def test_add_effect(self):
        nparams = len(self.model.parameters)
        self.model.add_effect(sncosmo.F99Dust(), 'host', 'rest')
        assert len(self.model.effects) == 2
        assert 'hostebv' in self.model.param_names
        assert len(self.model.parameters) == nparams + 1

    def test_set_and_get(self):
        """Test different methods for setting and getting parameters."""

        a = self.model.get('amplitude')
        self.model['amplitude'] = 2 * a   # test __setitem__
        assert self.model['amplitude'] == 2 * a   # test __getitem__

        self.model.update({'amplitude': 4 * a})  # test update
        assert self.model['amplitude'] == 4 * a


def test_effect_frame_free():
    """Test Model with PropagationEffect with a 'free' frame."""

    model = sncosmo.Model(source=flatsource(),
                          effects=[StepEffect(800., 20000.)],
                          effect_frames=['free'],
                          effect_names=['screen'])
    model.set(z=1.0, screenz=0.5, screenstepwave=10000.)

    # maxwave is set by the effect at z=0.5
    assert model.maxwave() == (1. + 0.5) * 20000.

    wave = np.array([12000., 13000., 14000., 14999., 15000., 16000.])
    assert_allclose(model.flux(0., wave), [0., 0., 0., 0., 0.5, 0.5])


def test_effect_phase_dependent():
    """Test Model with PropagationEffect with phase dependence"""

    model = sncosmo.Model(source=flatsource())
    flux = model.flux(50., 5000.).flatten()

    model_micro = sncosmo.Model(source=flatsource(),
                                effects=[AchromaticMicrolensing([40., 60.],
                                                                [10., 10.])],
                                effect_frames=['rest'],
                                effect_names=['micro'])
    assert_approx_equal(model_micro.flux(50., 5000.).flatten(), flux*10.)



#!/usr/bin/env python
# Licensed under a 3-clause BSD style license - see LICENSE.rst
import numpy
import os
import re

from setuptools import setup
from setuptools.extension import Extension

# Synchronize version from code.
VERSION = re.findall(r"__version__ = \"(.*?)\"",
                     open(os.path.join("sncosmo", "__init__.py")).read())[0]

# Cython extensions
extensions = [
    Extension(
        "sncosmo.salt2utils",
        sources=[os.path.join("sncosmo", "salt2utils.pyx")],
        include_dirs=[numpy.get_include()],
    )
]

setup(
    version=VERSION,
    ext_modules=extensions,
)



# Licensed under a 3-clause BSD style license - see LICENSES

from os.path import dirname, join

from numpy.testing import assert_allclose
import pytest

import sncosmo


def test_read_snana_ascii():
    fname = join(dirname(__file__), "data", "snana_ascii_example.dat")
    meta, tables = sncosmo.read_snana_ascii(fname, default_tablename="OBS")

    # test a few different types of metadata
    assert meta['SURVEY'] == 'DES'
    assert meta['FAKE'] == 0
    assert_allclose(meta['REDSHIFT_HELIO'], 0.3614)

    # only 1 table
    assert len(tables) == 1
    data = tables["OBS"]

    assert len(data) == 4  # 4 rows.
    assert len(data.colnames) == 13  # 13 columns.


def test_read_snana_fits():
    fname1 = join(dirname(__file__), "data", "snana_fits_example_head.fits")
    fname2 = join(dirname(__file__), "data", "snana_fits_example_phot.fits")
    sne = sncosmo.read_snana_fits(fname1, fname2)
    assert len(sne) == 2


def test_read_snana_simlib():
    fname = join(dirname(__file__), "data", "snana_simlib_example.dat")
    meta, obs_sets = sncosmo.read_snana_simlib(fname)
    assert len(obs_sets) == 2


def test_read_snana_simlib_noend():
    fname = join(dirname(__file__), "data", "snana_simlib_example_noend.dat")
    meta, obs_sets = sncosmo.read_snana_simlib(fname)
    assert len(obs_sets) == 2


def test_read_snana_simlib_doc():
    """Test when DOCANA header is present in simlib"""
    fname = join(dirname(__file__), "data", "snana_simlib_example_doc.dat")
    meta, _ = sncosmo.read_snana_simlib(fname)
    assert "DOCUMENTATION" in meta


def test_read_snana_simlib_coadd():
    """Test when co-added `ID*NEXPOSE` key is used."""
    fname = join(dirname(__file__), "data", "snana_simlib_example_coadd.dat")
    meta, obs_sets = sncosmo.read_snana_simlib(fname)
    assert "NEXPOSE" in obs_sets[0].colnames


def test_read_snana_simlib_invalid():
    """Test that we fails on simlib files with bad co-add columns."""
    fname = join(dirname(__file__), "data", "snana_simlib_example_invalid.dat")
    with pytest.raises(ValueError) as e_info:
        _ = sncosmo.read_snana_simlib(fname)
